# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
"""The raw JSON Web Token (JWT)."""

import copy
import datetime
import json

from typing import cast, Mapping, Set, List, Dict, Optional, Union, Any

from tink import core
from tink.jwt import _json_util
from tink.jwt import _jwt_error

_REGISTERED_NAMES = frozenset({'iss', 'sub', 'jti', 'aud', 'exp', 'nbf', 'iat'})

_MAX_TIMESTAMP_VALUE = 253402300799  # 31 Dec 9999, 23:59:59 GMT

Claim = Union[None, bool, int, float, str, List[Any], Dict[str, Any]]


def _from_datetime(t: datetime.datetime) -> int:
  if not t.tzinfo:
    raise _jwt_error.JwtInvalidError('datetime must have tzinfo')
  return int(t.timestamp())


def _to_datetime(timestamp: float) -> datetime.datetime:
  return datetime.datetime.fromtimestamp(timestamp, datetime.timezone.utc)


def _validate_custom_claim_name(name: str) -> None:
  if name in _REGISTERED_NAMES:
    raise _jwt_error.JwtInvalidError(
        'registered name %s cannot be custom claim name' % name)


class RawJwt:
  """An unencoded and unsigned JSON Web Token (JWT).

  It contains all payload claims and a subset of the headers. It does not
  contain any headers that depend on the key, such as "alg" or "kid", because
  these headers are chosen when the token is signed and encoded, and should not
  be chosen by the user. This ensures that the key can be changed without any
  changes to the user code.
  """

  def __new__(cls):
    raise core.TinkError('RawJwt cannot be instantiated directly.')

  def __init__(self, type_header: Optional[str], payload: Any) -> None:
    # No need to copy payload, because only create and from_json_payload
    # call this method.
    if not isinstance(payload, Dict):
      raise _jwt_error.JwtInvalidError('payload must be a dict')
    self._type_header = type_header
    self._payload = payload
    self._validate_string_claim('iss')
    self._validate_string_claim('sub')
    self._validate_string_claim('jti')
    self._validate_timestamp_claim('exp')
    self._validate_timestamp_claim('nbf')
    self._validate_timestamp_claim('iat')
    self._validate_audience_claim()

  def _validate_string_claim(self, name: str):
    if name in self._payload:
      if not isinstance(self._payload[name], str):
        raise _jwt_error.JwtInvalidError('claim %s must be a String' % name)

  def _validate_timestamp_claim(self, name: str):
    if name in self._payload:
      timestamp = self._payload[name]
      if not isinstance(timestamp, (int, float)):
        raise _jwt_error.JwtInvalidError('claim %s must be a Number' % name)
      if timestamp > _MAX_TIMESTAMP_VALUE or timestamp < 0:
        raise _jwt_error.JwtInvalidError(
            'timestamp of claim %s is out of range' % name)

  def _validate_audience_claim(self):
    """The 'aud' claim must either be a string or a list of strings."""
    if 'aud' in self._payload:
      audiences = self._payload['aud']
      if isinstance(audiences, str):
        return
      if not isinstance(audiences, list) or not audiences:
        raise _jwt_error.JwtInvalidError('audiences cannot be an empty list')
      if not all(isinstance(value, str) for value in audiences):
        raise _jwt_error.JwtInvalidError('audiences must only contain strings')

  def has_type_header(self) -> bool:
    return self._type_header is not None

  def type_header(self) -> str:
    if self._type_header is None:
      raise KeyError('type header is not set')
    return self._type_header

  def has_issuer(self) -> bool:
    return 'iss' in self._payload

  def issuer(self) -> str:
    return cast(str, self._payload['iss'])

  def has_subject(self) -> bool:
    return 'sub' in self._payload

  def subject(self) -> str:
    return cast(str, self._payload['sub'])

  def has_audiences(self) -> bool:
    return 'aud' in self._payload

  def audiences(self) -> List[str]:
    aud = self._payload['aud']
    if isinstance(aud, str):
      return [aud]
    return list(aud)

  def has_jwt_id(self) -> bool:
    return 'jti' in self._payload

  def jwt_id(self) -> str:
    return cast(str, self._payload['jti'])

  def has_expiration(self) -> bool:
    return 'exp' in self._payload

  def expiration(self) -> datetime.datetime:
    return _to_datetime(self._payload['exp'])

  def has_not_before(self) -> bool:
    return 'nbf' in self._payload

  def not_before(self) -> datetime.datetime:
    return _to_datetime(self._payload['nbf'])

  def has_issued_at(self) -> bool:
    return 'iat' in self._payload

  def issued_at(self) -> datetime.datetime:
    return _to_datetime(self._payload['iat'])

  def custom_claim_names(self) -> Set[str]:
    return {n for n in self._payload.keys() if n not in _REGISTERED_NAMES}

  def custom_claim(self, name: str) -> Claim:
    _validate_custom_claim_name(name)
    value = self._payload[name]
    if isinstance(value, (list, dict)):
      return copy.deepcopy(value)
    else:
      return value

  def json_payload(self) -> str:
    """Returns the payload encoded as JSON string."""
    return _json_util.json_dumps(self._payload)

  @classmethod
  def create(cls,
             *,
             type_header: Optional[str] = None,
             issuer: Optional[str] = None,
             subject: Optional[str] = None,
             audience: Optional[str] = None,
             audiences: Optional[List[str]] = None,
             jwt_id: Optional[str] = None,
             expiration: Optional[datetime.datetime] = None,
             without_expiration: Optional[bool] = None,
             not_before: Optional[datetime.datetime] = None,
             issued_at: Optional[datetime.datetime] = None,
             custom_claims: Optional[Mapping[str, Claim]] = None) -> 'RawJwt':
    """Create a new RawJwt instance."""
    if not expiration and not without_expiration:
      raise ValueError('either expiration or without_expiration must be set')
    if expiration and without_expiration:
      raise ValueError(
          'expiration and without_expiration cannot be set at the same time')
    if audience is not None and audiences is not None:
      raise _jwt_error.JwtInvalidError(
          'audience and audiences cannot be set at the same time')
    payload: Dict[str, Any] = {}
    if issuer:
      payload['iss'] = issuer
    if subject:
      payload['sub'] = subject
    if jwt_id is not None:
      payload['jti'] = jwt_id
    if audience is not None:
      payload['aud'] = audience
    if audiences is not None:
      payload['aud'] = copy.copy(audiences)
    if expiration:
      payload['exp'] = _from_datetime(expiration)
    if not_before:
      payload['nbf'] = _from_datetime(not_before)
    if issued_at:
      payload['iat'] = _from_datetime(issued_at)
    if custom_claims:
      for name, value in custom_claims.items():
        _validate_custom_claim_name(name)
        if not isinstance(name, str):
          raise _jwt_error.JwtInvalidError('claim name must be Text')
        if (value is None or isinstance(value, (bool, int, float, str))):
          payload[name] = value
        elif isinstance(value, list):
          payload[name] = json.loads(json.dumps(value))
        elif isinstance(value, dict):
          payload[name] = json.loads(json.dumps(value))
        else:
          raise _jwt_error.JwtInvalidError('claim %s has unknown type' % name)
    raw_jwt = object.__new__(cls)
    raw_jwt.__init__(type_header, payload)
    return raw_jwt

  @classmethod
  def _from_json(cls, type_header: Optional[str], payload: str) -> 'RawJwt':
    """Creates a RawJwt from payload encoded as JSON string."""
    raw_jwt = object.__new__(cls)
    raw_jwt.__init__(type_header, _json_util.json_loads(payload))
    return raw_jwt


def new_raw_jwt(*,
                type_header: Optional[str] = None,
                issuer: Optional[str] = None,
                subject: Optional[str] = None,
                audience: Optional[str] = None,
                audiences: Optional[List[str]] = None,
                jwt_id: Optional[str] = None,
                expiration: Optional[datetime.datetime] = None,
                without_expiration: bool = False,
                not_before: Optional[datetime.datetime] = None,
                issued_at: Optional[datetime.datetime] = None,
                custom_claims: Optional[Mapping[str, Claim]] = None) -> RawJwt:
  """Creates a new RawJwt."""
  return RawJwt.create(
      type_header=type_header,
      issuer=issuer,
      subject=subject,
      audience=audience,
      audiences=audiences,
      jwt_id=jwt_id,
      expiration=expiration,
      without_expiration=without_expiration,
      not_before=not_before,
      issued_at=issued_at,
      custom_claims=custom_claims)


def raw_jwt_from_json(type_header: Optional[str], payload: str) -> RawJwt:
  """Internal function used to verify JWT token."""
  return RawJwt._from_json(type_header, payload)  # pylint: disable=protected-access



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for tink.python.tink.cc.pybind.py_key_manager."""

from typing import cast
from absl.testing import absltest
from absl.testing import parameterized
from tink.proto import aes_eax_pb2
from tink.proto import aes_siv_pb2
from tink.proto import common_pb2
from tink.proto import ecdsa_pb2
from tink.proto import ecies_aead_hkdf_pb2
from tink.proto import hmac_pb2
from tink.proto import hmac_prf_pb2
from tink.proto import hpke_pb2
from tink.proto import jwt_ecdsa_pb2
from tink.proto import jwt_hmac_pb2
from tink.proto import tink_pb2
from tink import aead
from tink import hybrid
from tink.cc.pybind import tink_bindings


def setUpModule():
  tink_bindings.register()
  tink_bindings.register_jwt()
  tink_bindings.register_hpke()


class AeadKeyManagerTest(absltest.TestCase):

  def setUp(self):
    super().setUp()
    self.key_manager = tink_bindings.AeadKeyManager.from_cc_registry(
        'type.googleapis.com/google.crypto.tink.AesEaxKey')

  def new_aes_eax_key_template(self, iv_size, key_size):
    key_format = aes_eax_pb2.AesEaxKeyFormat()
    key_format.params.iv_size = iv_size
    key_format.key_size = key_size
    key_template = tink_pb2.KeyTemplate()
    key_template.type_url = 'type.googleapis.com/google.crypto.tink.AesEaxKey'
    key_template.value = key_format.SerializeToString()
    return key_template.SerializeToString()

  def test_key_type(self):
    self.assertEqual(self.key_manager.key_type(),
                     'type.googleapis.com/google.crypto.tink.AesEaxKey')

  def test_new_key_data(self):
    key_template = self.new_aes_eax_key_template(12, 16)
    serialized_key_data = self.key_manager.new_key_data(key_template)
    key_data = tink_pb2.KeyData.FromString(serialized_key_data)
    self.assertEqual(key_data.type_url, self.key_manager.key_type())
    self.assertEqual(key_data.key_material_type, tink_pb2.KeyData.SYMMETRIC)
    key = aes_eax_pb2.AesEaxKey.FromString(key_data.value)
    self.assertEqual(key.version, 0)
    self.assertEqual(key.params.iv_size, 12)
    self.assertLen(key.key_value, 16)

  def test_invalid_params_raise_exception(self):
    key_template = self.new_aes_eax_key_template(9, 16)
    with self.assertRaises(tink_bindings.PythonTinkException):
      self.key_manager.new_key_data(key_template)

  def test_encrypt_decrypt(self):
    key_template = self.new_aes_eax_key_template(12, 16)
    key_data = self.key_manager.new_key_data(key_template)

    primitive = self.key_manager.primitive(key_data)
    plaintext = b'plaintext'
    associated_data = b'associated_data'
    ciphertext = primitive.encrypt(plaintext, associated_data)
    self.assertEqual(primitive.decrypt(ciphertext, associated_data), plaintext)


class DeterministicAeadKeyManagerTest(absltest.TestCase):

  def setUp(self):
    super().setUp()
    daead_key_manager = tink_bindings.DeterministicAeadKeyManager
    self.key_manager = daead_key_manager.from_cc_registry(
        'type.googleapis.com/google.crypto.tink.AesSivKey')

  def new_aes_siv_key_template(self, key_size):
    key_format = aes_siv_pb2.AesSivKeyFormat()
    key_format.key_size = key_size
    key_template = tink_pb2.KeyTemplate()
    key_template.type_url = 'type.googleapis.com/google.crypto.tink.AesSivKey'
    key_template.value = key_format.SerializeToString()
    return key_template.SerializeToString()

  def test_key_type(self):
    self.assertEqual(self.key_manager.key_type(),
                     'type.googleapis.com/google.crypto.tink.AesSivKey')

  def test_new_key_data(self):
    key_template = self.new_aes_siv_key_template(64)
    key_data = tink_pb2.KeyData.FromString(
        self.key_manager.new_key_data(key_template))
    self.assertEqual(key_data.type_url, self.key_manager.key_type())
    self.assertEqual(key_data.key_material_type, tink_pb2.KeyData.SYMMETRIC)
    key = aes_siv_pb2.AesSivKey.FromString(key_data.value)
    self.assertEqual(key.version, 0)
    self.assertLen(key.key_value, 64)

  def test_invalid_params_raise_exception(self):
    key_template = self.new_aes_siv_key_template(65)
    with self.assertRaises(tink_bindings.PythonTinkException):
      self.key_manager.new_key_data(key_template)

  def test_encrypt_decrypt(self):
    key_template = self.new_aes_siv_key_template(64)
    key_data = self.key_manager.new_key_data(key_template)

    primitive = self.key_manager.primitive(key_data)
    plaintext = b'plaintext'
    associated_data = b'associated_data'
    ciphertext = primitive.encrypt_deterministically(plaintext, associated_data)
    self.assertEqual(
        primitive.decrypt_deterministically(ciphertext, associated_data),
        plaintext)


class HybridKeyManagerTest(absltest.TestCase):

  def hybrid_decrypt_key_manager(self):
    return tink_bindings.HybridDecryptKeyManager.from_cc_registry(
        'type.googleapis.com/google.crypto.tink.EciesAeadHkdfPrivateKey')

  def hybrid_encrypt_key_manager(self):
    return tink_bindings.HybridEncryptKeyManager.from_cc_registry(
        'type.googleapis.com/google.crypto.tink.EciesAeadHkdfPublicKey')

  def test_new_key_data(self):
    key_manager = self.hybrid_decrypt_key_manager()
    key_data = tink_pb2.KeyData.FromString(
        key_manager.new_key_data(
            hybrid.hybrid_key_templates.ECIES_P256_HKDF_HMAC_SHA256_AES128_GCM
            .SerializeToString()))
    self.assertEqual(key_data.type_url, key_manager.key_type())
    self.assertEqual(key_data.key_material_type,
                     tink_pb2.KeyData.ASYMMETRIC_PRIVATE)
    key = ecies_aead_hkdf_pb2.EciesAeadHkdfPrivateKey.FromString(key_data.value)
    self.assertLen(key.key_value, 32)
    self.assertEqual(key.public_key.params.kem_params.curve_type,
                     common_pb2.NIST_P256)

  def test_new_key_data_invalid_params_raise_exception(self):
    with self.assertRaisesRegex(tink_bindings.PythonTinkException,
                                'Unsupported elliptic curve'):
      self.hybrid_decrypt_key_manager().new_key_data(
          hybrid.hybrid_key_templates.create_ecies_aead_hkdf_key_template(
              curve_type=cast(common_pb2.EllipticCurveType, 100),  # invalid
              ec_point_format=common_pb2.UNCOMPRESSED,
              hash_type=common_pb2.SHA256,
              dem_key_template=aead.aead_key_templates.AES128_GCM)
          .SerializeToString())

  def test_encrypt_decrypt(self):
    decrypt_key_manager = self.hybrid_decrypt_key_manager()
    encrypt_key_manager = self.hybrid_encrypt_key_manager()
    key_data = decrypt_key_manager.new_key_data(
        hybrid.hybrid_key_templates.ECIES_P256_HKDF_HMAC_SHA256_AES128_GCM
        .SerializeToString())
    public_key_data = decrypt_key_manager.public_key_data(key_data)
    hybrid_encrypt = encrypt_key_manager.primitive(public_key_data)
    ciphertext = hybrid_encrypt.encrypt(b'some plaintext', b'some context info')
    hybrid_decrypt = decrypt_key_manager.primitive(key_data)
    self.assertEqual(hybrid_decrypt.decrypt(ciphertext, b'some context info'),
                     b'some plaintext')

  def test_decrypt_fails(self):
    decrypt_key_manager = self.hybrid_decrypt_key_manager()
    key_data = decrypt_key_manager.new_key_data(
        hybrid.hybrid_key_templates.ECIES_P256_HKDF_HMAC_SHA256_AES128_GCM
        .SerializeToString())
    hybrid_decrypt = decrypt_key_manager.primitive(key_data)
    with self.assertRaisesRegex(tink_bindings.PythonTinkException,
                                'ciphertext too short'):
      hybrid_decrypt.decrypt(b'bad ciphertext', b'some context info')


class HpkeKeyManagerTest(parameterized.TestCase):

  def hybrid_decrypt_key_manager(self):
    return tink_bindings.HybridDecryptKeyManager.from_cc_registry(
        'type.googleapis.com/google.crypto.tink.HpkePrivateKey')

  def hybrid_encrypt_key_manager(self):
    return tink_bindings.HybridEncryptKeyManager.from_cc_registry(
        'type.googleapis.com/google.crypto.tink.HpkePublicKey')

  def test_new_key_data(self):
    key_manager = self.hybrid_decrypt_key_manager()
    key_data = tink_pb2.KeyData.FromString(
        key_manager.new_key_data(
            hybrid.hybrid_key_templates
            .DHKEM_X25519_HKDF_SHA256_HKDF_SHA256_AES_128_GCM.SerializeToString(
            )))
    self.assertEqual(key_data.type_url, key_manager.key_type())
    self.assertEqual(key_data.key_material_type,
                     tink_pb2.KeyData.ASYMMETRIC_PRIVATE)
    key = hpke_pb2.HpkePrivateKey.FromString(key_data.value)
    self.assertLen(key.private_key, 32)  # HPKE 'Nsk' parameter length  = 32
    self.assertEqual(key.public_key.params.kem,
                     hpke_pb2.DHKEM_X25519_HKDF_SHA256)
    self.assertEqual(key.public_key.params.kdf, hpke_pb2.HKDF_SHA256)
    self.assertEqual(key.public_key.params.aead, hpke_pb2.AES_128_GCM)

  def test_new_key_data_invalid_kem_raise_exception(self):
    with self.assertRaisesRegex(tink_bindings.PythonTinkException,
                                'Invalid KEM param.'):
      self.hybrid_decrypt_key_manager().new_key_data(
          hybrid.hybrid_key_templates._create_hpke_key_template(
              hpke_kem=hpke_pb2.KEM_UNKNOWN,
              hpke_kdf=hpke_pb2.HKDF_SHA256,
              hpke_aead=hpke_pb2.AES_128_GCM,
              output_prefix_type=tink_pb2.TINK).SerializeToString())

  def test_new_key_data_invalid_kdf_raise_exception(self):
    with self.assertRaisesRegex(tink_bindings.PythonTinkException,
                                'Invalid KDF param.'):
      self.hybrid_decrypt_key_manager().new_key_data(
          hybrid.hybrid_key_templates._create_hpke_key_template(
              hpke_kem=hpke_pb2.DHKEM_X25519_HKDF_SHA256,
              hpke_kdf=hpke_pb2.KDF_UNKNOWN,
              hpke_aead=hpke_pb2.AES_128_GCM,
              output_prefix_type=tink_pb2.TINK).SerializeToString())

  def test_new_key_data_invalid_aead_raise_exception(self):
    with self.assertRaisesRegex(tink_bindings.PythonTinkException,
                                'Invalid AEAD param.'):
      self.hybrid_decrypt_key_manager().new_key_data(
          hybrid.hybrid_key_templates._create_hpke_key_template(
              hpke_kem=hpke_pb2.DHKEM_X25519_HKDF_SHA256,
              hpke_kdf=hpke_pb2.HKDF_SHA256,
              hpke_aead=hpke_pb2.AEAD_UNKNOWN,
              output_prefix_type=tink_pb2.TINK).SerializeToString())

  def test_encrypt_decrypt(self):
    decrypt_key_manager = self.hybrid_decrypt_key_manager()
    encrypt_key_manager = self.hybrid_encrypt_key_manager()
    key_data = decrypt_key_manager.new_key_data(
        hybrid.hybrid_key_templates
        .DHKEM_X25519_HKDF_SHA256_HKDF_SHA256_AES_128_GCM.SerializeToString())
    public_key_data = decrypt_key_manager.public_key_data(key_data)
    hybrid_encrypt = encrypt_key_manager.primitive(public_key_data)
    ciphertext = hybrid_encrypt.encrypt(b'some plaintext', b'some context info')
    hybrid_decrypt = decrypt_key_manager.primitive(key_data)
    self.assertEqual(
        hybrid_decrypt.decrypt(ciphertext, b'some context info'),
        b'some plaintext')

  @parameterized.parameters(
      [tink_pb2.TINK, tink_pb2.RAW, tink_pb2.CRUNCHY, tink_pb2.LEGACY])
  def test_encrypt_decrypt_by_prefix(self, prefix):
    decrypt_key_manager = self.hybrid_decrypt_key_manager()
    encrypt_key_manager = self.hybrid_encrypt_key_manager()
    key_data = decrypt_key_manager.new_key_data(
        hybrid.hybrid_key_templates._create_hpke_key_template(
            hpke_kem=hpke_pb2.DHKEM_X25519_HKDF_SHA256,
            hpke_kdf=hpke_pb2.HKDF_SHA256,
            hpke_aead=hpke_pb2.AES_128_GCM,
            output_prefix_type=prefix).SerializeToString())
    public_key_data = decrypt_key_manager.public_key_data(key_data)
    hybrid_encrypt = encrypt_key_manager.primitive(public_key_data)
    ciphertext = hybrid_encrypt.encrypt(b'some plaintext', b'some context info')
    hybrid_decrypt = decrypt_key_manager.primitive(key_data)
    self.assertEqual(
        hybrid_decrypt.decrypt(ciphertext, b'some context info'),
        b'some plaintext')

  def test_decrypt_fails(self):
    decrypt_key_manager = self.hybrid_decrypt_key_manager()
    key_data = decrypt_key_manager.new_key_data(
        hybrid.hybrid_key_templates
        .DHKEM_X25519_HKDF_SHA256_HKDF_SHA256_AES_128_GCM.SerializeToString())
    hybrid_decrypt = decrypt_key_manager.primitive(key_data)
    with self.assertRaisesRegex(tink_bindings.PythonTinkException,
                                'Ciphertext is too short.'):
      hybrid_decrypt.decrypt(b'bad ciphertext', b'some context info')


class MacKeyManagerTest(absltest.TestCase):

  def setUp(self):
    super().setUp()
    self.key_manager = tink_bindings.MacKeyManager.from_cc_registry(
        'type.googleapis.com/google.crypto.tink.HmacKey')

  def new_hmac_key_template(self, hash_type, tag_size, key_size):
    key_format = hmac_pb2.HmacKeyFormat()
    key_format.params.hash = hash_type
    key_format.params.tag_size = tag_size
    key_format.key_size = key_size
    key_template = tink_pb2.KeyTemplate()
    key_template.type_url = 'type.googleapis.com/google.crypto.tink.HmacKey'
    key_template.value = key_format.SerializeToString()
    return key_template.SerializeToString()

  def test_key_type(self):
    self.assertEqual(self.key_manager.key_type(),
                     'type.googleapis.com/google.crypto.tink.HmacKey')

  def test_new_key_data(self):
    key_template = self.new_hmac_key_template(common_pb2.SHA256, 24, 16)
    key_data = tink_pb2.KeyData.FromString(
        self.key_manager.new_key_data(key_template))
    self.assertEqual(key_data.type_url, self.key_manager.key_type())
    key = hmac_pb2.HmacKey.FromString(key_data.value)
    self.assertEqual(key.version, 0)
    self.assertEqual(key.params.hash, common_pb2.SHA256)
    self.assertEqual(key.params.tag_size, 24)
    self.assertLen(key.key_value, 16)

  def test_invalid_params_raise_exception(self):
    key_template = self.new_hmac_key_template(common_pb2.SHA256, 9, 16)
    with self.assertRaises(tink_bindings.PythonTinkException):
      self.key_manager.new_key_data(key_template)

  def test_mac_success(self):
    mac = self.key_manager.primitive(
        self.key_manager.new_key_data(
            self.new_hmac_key_template(common_pb2.SHA256, 24, 16)))
    data = b'data'
    tag = mac.compute_mac(data)
    self.assertLen(tag, 24)
    # No exception raised.
    mac.verify_mac(tag, data)

  def test_mac_wrong(self):
    mac = self.key_manager.primitive(
        self.key_manager.new_key_data(
            self.new_hmac_key_template(common_pb2.SHA256, 16, 16)))
    with self.assertRaisesRegex(tink_bindings.PythonTinkException,
                                'verification failed'):
      mac.verify_mac(b'0123456789ABCDEF', b'data')


class JwtMacKeyManagerTest(absltest.TestCase):

  def setUp(self):
    super().setUp()
    self.key_manager = tink_bindings.MacKeyManager.from_cc_registry(
        'type.googleapis.com/google.crypto.tink.JwtHmacKey')

  def new_jwt_hmac_key_template(self, algorithm, key_size):
    key_format = jwt_hmac_pb2.JwtHmacKeyFormat()
    key_format.algorithm = algorithm
    key_format.key_size = key_size
    key_template = tink_pb2.KeyTemplate()
    key_template.type_url = 'type.googleapis.com/google.crypto.tink.JwtHmacKey'
    key_template.value = key_format.SerializeToString()
    return key_template.SerializeToString()

  def test_key_type(self):
    self.assertEqual(self.key_manager.key_type(),
                     'type.googleapis.com/google.crypto.tink.JwtHmacKey')

  def test_new_key_data(self):
    key_template = self.new_jwt_hmac_key_template(jwt_hmac_pb2.HS256, 32)
    key_data = tink_pb2.KeyData.FromString(
        self.key_manager.new_key_data(key_template))
    self.assertEqual(key_data.type_url, self.key_manager.key_type())
    key = jwt_hmac_pb2.JwtHmacKey.FromString(key_data.value)
    self.assertEqual(key.version, 0)
    self.assertEqual(key.algorithm, jwt_hmac_pb2.HS256)
    self.assertLen(key.key_value, 32)

  def test_too_short_key_size_raises_exception(self):
    key_template = self.new_jwt_hmac_key_template(jwt_hmac_pb2.HS256, 31)
    with self.assertRaises(tink_bindings.PythonTinkException):
      self.key_manager.new_key_data(key_template)

  def test_mac_success(self):
    mac = self.key_manager.primitive(
        self.key_manager.new_key_data(
            self.new_jwt_hmac_key_template(jwt_hmac_pb2.HS256, 32)))
    data = b'data'
    tag = mac.compute_mac(data)
    self.assertLen(tag, 32)
    # No exception raised.
    mac.verify_mac(tag, data)

  def test_mac_wrong(self):
    mac = self.key_manager.primitive(
        self.key_manager.new_key_data(
            self.new_jwt_hmac_key_template(jwt_hmac_pb2.HS256, 32)))
    with self.assertRaisesRegex(tink_bindings.PythonTinkException,
                                'verification failed'):
      mac.verify_mac(b'0123456789ABCDEF0123456789ABCDEF', b'data')


class PrfKeyManagerTest(absltest.TestCase):

  def setUp(self):
    super().setUp()
    self.key_manager = tink_bindings.PrfKeyManager.from_cc_registry(
        'type.googleapis.com/google.crypto.tink.HmacPrfKey')

  def new_hmac_prf_key_template(self, hash_type, key_size):
    key_format = hmac_prf_pb2.HmacPrfKeyFormat()
    key_format.params.hash = hash_type
    key_format.key_size = key_size
    key_format.version = 0
    key_template = tink_pb2.KeyTemplate()
    key_template.type_url = 'type.googleapis.com/google.crypto.tink.HmacPrfKey'
    key_template.value = key_format.SerializeToString()
    return key_template.SerializeToString()

  def test_key_type(self):
    self.assertEqual(self.key_manager.key_type(),
                     'type.googleapis.com/google.crypto.tink.HmacPrfKey')

  def test_new_key_data(self):
    key_template = self.new_hmac_prf_key_template(
        hash_type=common_pb2.SHA256, key_size=16)
    key_data = tink_pb2.KeyData.FromString(
        self.key_manager.new_key_data(key_template))
    self.assertEqual(key_data.type_url, self.key_manager.key_type())
    key = hmac_pb2.HmacKey.FromString(key_data.value)
    self.assertEqual(key.version, 0)
    self.assertEqual(key.params.hash, common_pb2.SHA256)
    self.assertLen(key.key_value, 16)

  def test_invalid_params_raise_exception(self):
    key_template = self.new_hmac_prf_key_template(
        hash_type=common_pb2.SHA256, key_size=7)
    with self.assertRaises(tink_bindings.PythonTinkException):
      self.key_manager.new_key_data(key_template)

  def test_prf_success(self):
    prf = self.key_manager.primitive(
        self.key_manager.new_key_data(
            self.new_hmac_prf_key_template(
                hash_type=common_pb2.SHA256, key_size=16)))
    output = prf.compute(b'input_data', output_length=31)
    self.assertLen(output, 31)
    self.assertEqual(prf.compute(b'input_data', output_length=31), output)

  def test_prf_bad_output_length(self):
    prf = self.key_manager.primitive(
        self.key_manager.new_key_data(
            self.new_hmac_prf_key_template(
                hash_type=common_pb2.SHA256, key_size=16)))
    with self.assertRaises(tink_bindings.PythonTinkException):
      _ = prf.compute(b'input_data', output_length=12345)


class PublicKeySignVerifyKeyManagerTest(absltest.TestCase):

  def setUp(self):
    super().setUp()
    public_key_verify_manager = tink_bindings.PublicKeyVerifyKeyManager
    self.key_manager_verify = public_key_verify_manager.from_cc_registry(
        'type.googleapis.com/google.crypto.tink.EcdsaPublicKey')
    public_key_sign_manager = tink_bindings.PublicKeySignKeyManager
    self.key_manager_sign = public_key_sign_manager.from_cc_registry(
        'type.googleapis.com/google.crypto.tink.EcdsaPrivateKey')

  def new_ecdsa_key_template(self, hash_type, curve_type, encoding,
                             public=False):
    params = ecdsa_pb2.EcdsaParams(
        hash_type=hash_type, curve=curve_type, encoding=encoding)
    key_format = ecdsa_pb2.EcdsaKeyFormat(params=params)
    key_template = tink_pb2.KeyTemplate()
    if public:
      append = 'EcdsaPublicKey'
    else:
      append = 'EcdsaPrivateKey'
    key_template.type_url = 'type.googleapis.com/google.crypto.tink.' + append

    key_template.value = key_format.SerializeToString()
    return key_template.SerializeToString()

  def test_key_type_sign(self):
    self.assertEqual(self.key_manager_sign.key_type(),
                     'type.googleapis.com/google.crypto.tink.EcdsaPrivateKey')

  def test_key_type_verify(self):
    self.assertEqual(self.key_manager_verify.key_type(),
                     'type.googleapis.com/google.crypto.tink.EcdsaPublicKey')

  def test_new_key_data_sign(self):
    key_template = self.new_ecdsa_key_template(
        common_pb2.SHA256, common_pb2.NIST_P256, ecdsa_pb2.DER)
    key_data = tink_pb2.KeyData.FromString(
        self.key_manager_sign.new_key_data(key_template))
    self.assertEqual(key_data.type_url, self.key_manager_sign.key_type())
    key = ecdsa_pb2.EcdsaPrivateKey.FromString(key_data.value)
    public_key = key.public_key
    self.assertEqual(key.version, 0)
    self.assertEqual(public_key.version, 0)
    self.assertEqual(public_key.params.hash_type, common_pb2.SHA256)
    self.assertEqual(public_key.params.curve, common_pb2.NIST_P256)
    self.assertEqual(public_key.params.encoding, ecdsa_pb2.DER)
    self.assertLen(key.key_value, 32)

  def test_new_key_data_verify(self):
    key_template = self.new_ecdsa_key_template(
        common_pb2.SHA256, common_pb2.NIST_P256, ecdsa_pb2.DER, True)
    with self.assertRaisesRegex(tink_bindings.PythonTinkException,
                                'not supported'):
      self.key_manager_verify.new_key_data(key_template)

  def test_signature_success(self):
    priv_key = self.key_manager_sign.new_key_data(
        self.new_ecdsa_key_template(common_pb2.SHA256, common_pb2.NIST_P256,
                                    ecdsa_pb2.DER))
    pub_key = self.key_manager_sign.public_key_data(priv_key)

    verifier = self.key_manager_verify.primitive(pub_key)
    signer = self.key_manager_sign.primitive(priv_key)

    data = b'data'
    signature = signer.sign(data)

    # Starts with a DER sequence
    self.assertEqual(bytearray(signature)[0], 0x30)

    verifier.verify(signature, data)

  def test_signature_fails(self):
    key_template = self.new_ecdsa_key_template(
        common_pb2.SHA256, common_pb2.NIST_P256, ecdsa_pb2.DER, False)
    priv_key = self.key_manager_sign.new_key_data(key_template)
    pub_key = self.key_manager_sign.public_key_data(priv_key)

    signer = self.key_manager_sign.primitive(priv_key)
    verifier = self.key_manager_verify.primitive(pub_key)

    data = b'data'
    signature = signer.sign(data)

    with self.assertRaisesRegex(tink_bindings.PythonTinkException,
                                'Signature is not valid'):
      verifier.verify(signature, b'wrongdata')

    with self.assertRaisesRegex(tink_bindings.PythonTinkException,
                                'Signature is not valid'):
      verifier.verify(b'wrongsignature', data)


class JwtPublicKeySignVerifyKeyManagerTest(absltest.TestCase):

  def setUp(self):
    super().setUp()
    public_key_verify_manager = tink_bindings.PublicKeyVerifyKeyManager
    self.key_manager_verify = public_key_verify_manager.from_cc_registry(
        'type.googleapis.com/google.crypto.tink.JwtEcdsaPublicKey')
    public_key_sign_manager = tink_bindings.PublicKeySignKeyManager
    self.key_manager_sign = public_key_sign_manager.from_cc_registry(
        'type.googleapis.com/google.crypto.tink.JwtEcdsaPrivateKey')

  def test_new_key_data_verify_fails(self):
    key_format = jwt_ecdsa_pb2.JwtEcdsaKeyFormat(algorithm=jwt_ecdsa_pb2.ES256)
    key_template = tink_pb2.KeyTemplate()
    key_template.type_url = (
        'type.googleapis.com/google.crypto.tink.JwtEcdsaPublicKey')
    key_template.value = key_format.SerializeToString()
    with self.assertRaisesRegex(tink_bindings.PythonTinkException,
                                'not supported'):
      self.key_manager_verify.new_key_data(key_template.SerializeToString())

  def test_signature_success(self):
    key_format = jwt_ecdsa_pb2.JwtEcdsaKeyFormat(algorithm=jwt_ecdsa_pb2.ES256)
    key_template = tink_pb2.KeyTemplate()
    key_template.type_url = (
        'type.googleapis.com/google.crypto.tink.JwtEcdsaPrivateKey')
    key_template.value = key_format.SerializeToString()

    priv_key = self.key_manager_sign.new_key_data(
        key_template.SerializeToString())
    pub_key = self.key_manager_sign.public_key_data(priv_key)

    verifier = self.key_manager_verify.primitive(pub_key)
    signer = self.key_manager_sign.primitive(priv_key)

    data = b'data'
    signature = signer.sign(data)
    verifier.verify(signature, data)

    with self.assertRaises(tink_bindings.PythonTinkException):
      verifier.verify(signature, b'wrongdata')

    with self.assertRaises(tink_bindings.PythonTinkException):
      verifier.verify(b'wrongsignature', data)


if __name__ == '__main__':
  absltest.main()



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""A client for AWS KMS."""

import binascii
import configparser
import re
from typing import Any, Dict, Optional, Tuple

import boto3
from botocore import exceptions

import tink
from tink import aead


AWS_KEYURI_PREFIX = 'aws-kms://'


def _encryption_context(associated_data: bytes) -> Dict[str, str]:
  if associated_data:
    hex_associated_data = binascii.hexlify(associated_data).decode('utf-8')
    return {'associatedData': hex_associated_data}
  else:
    return dict()


class _AwsKmsAead(aead.Aead):
  """Implements the Aead interface for AWS KMS."""

  def __init__(self, client: Any, key_arn: str) -> None:
    self.client = client
    self.key_arn = key_arn

  def encrypt(self, plaintext: bytes, associated_data: bytes) -> bytes:
    try:
      response = self.client.encrypt(
          KeyId=self.key_arn,
          Plaintext=plaintext,
          EncryptionContext=_encryption_context(associated_data),
      )
      return response['CiphertextBlob']
    except exceptions.ClientError as e:
      raise tink.TinkError(e)

  def decrypt(self, ciphertext: bytes, associated_data: bytes) -> bytes:
    try:
      response = self.client.decrypt(
          KeyId=self.key_arn,
          CiphertextBlob=ciphertext,
          EncryptionContext=_encryption_context(associated_data),
      )
      return response['Plaintext']
    except exceptions.ClientError as e:
      raise tink.TinkError(e)


def _has_aws_key_uri_format(key_uri: str) -> bool:
  match = re.match('aws-kms://arn:aws:kms:([a-z0-9-]+):', key_uri)
  return match is not None


def _key_uri_to_key_arn(key_uri: str) -> str:
  if not key_uri.startswith(AWS_KEYURI_PREFIX):
    raise tink.TinkError('invalid key URI')
  return key_uri[len(AWS_KEYURI_PREFIX) :]


class _KmsClient(tink.KmsClient):
  """KMS client returned by new_client."""

  def __init__(self, boto3_client: Any, key_uri: Optional[str]):
    if not key_uri:
      self._key_uri = None
    else:
      if not _has_aws_key_uri_format(key_uri):
        raise tink.TinkError('invalid key URI')
      self._key_uri = key_uri
    self._boto3_client = boto3_client

  def does_support(self, key_uri: str) -> bool:
    if not _has_aws_key_uri_format(key_uri):
      return False
    if not self._key_uri:
      return True
    return key_uri == self._key_uri

  def get_aead(self, key_uri: str) -> aead.Aead:
    if not self.does_support(key_uri):
      if self._key_uri:
        raise tink.TinkError(
            'This client is bound to %s and cannot use key %s' %
            (self._key_uri, key_uri))
      raise tink.TinkError(
          'This client does not support key %s' % key_uri)
    return _AwsKmsAead(self._boto3_client, _key_uri_to_key_arn(key_uri))


def new_client(
    *, boto3_client: Any, key_uri: Optional[str] = None
) -> tink.KmsClient:
  """Creates a new Tink KmsClient from a boto3 client."""
  return _KmsClient(boto3_client, key_uri)


def _parse_config(config_path: str) -> Tuple[str, str]:
  """Returns ('aws_access_key_id', 'aws_secret_access_key') from a config."""
  config = configparser.ConfigParser()
  config.read(config_path)
  if 'default' not in config:
    raise ValueError('invalid config: default not found')
  default = config['default']
  if 'aws_access_key_id' not in default:
    raise ValueError('invalid config: aws_access_key_id not found')
  aws_access_key_id = default['aws_access_key_id']
  if 'aws_secret_access_key' not in default:
    raise ValueError('invalid config: aws_secret_access_key not found')
  aws_secret_access_key = default['aws_secret_access_key']
  return (aws_access_key_id, aws_secret_access_key)


def _get_region_from_key_arn(key_arn: str) -> str:
  # An AWS key ARN is of the form
  # arn:aws:kms:us-west-2:111122223333:key/1234abcd-12ab-34cd-56ef-1234567890ab.
  key_arn_parts = key_arn.split(':')
  if len(key_arn_parts) < 6:
    raise tink.TinkError('invalid key id')
  return key_arn_parts[3]


class AwsKmsClient(tink.KmsClient):
  """Basic AWS client for AEAD."""

  def __init__(self, key_uri: Optional[str], credentials_path: Optional[str]):
    """Creates a new AwsKmsClient that is bound to the key specified in 'key_uri'.

    For more information on credentials and in which order they are loaded see
    https://boto3.amazonaws.com/v1/documentation/api/latest/guide/configuration.html.

    Args:
      key_uri: The URI of the key the client should be bound to. If it is None
          or empty, then the client is not bound to any particular key.
      credentials_path: Path to the file with the access credentials. If it is
          None or empty, then default credentials will be used.

    Raises:
      ValueError: If the path or filename of the credentials is invalid.
      TinkError: If the key uri is not valid.
    """
    if not key_uri:
      self._key_uri = None
    else:
      if not _has_aws_key_uri_format(key_uri):
        raise tink.TinkError('invalid key URI')
      self._key_uri = key_uri
    if not credentials_path:
      self._aws_access_key_id = None
      self._aws_secret_access_key = None
    else:
      aws_access_key_id, aws_secret_access_key = _parse_config(credentials_path)
      self._aws_access_key_id = aws_access_key_id
      self._aws_secret_access_key = aws_secret_access_key

  def does_support(self, key_uri: str) -> bool:
    """Returns true if this client supports KMS key specified in 'key_uri'.

    Args:
      key_uri: Text, URI of the key to be checked.

    Returns: A boolean value which is true if the key is supported and false
      otherwise.
    """
    if not _has_aws_key_uri_format(key_uri):
      return False
    if not self._key_uri:
      return True
    return key_uri == self._key_uri

  def get_aead(self, key_uri: str) -> aead.Aead:
    """Returns an Aead-primitive backed by KMS key specified by 'key_uri'.

    Args:
      key_uri: Text, URI of the key which should be used.

    Returns:
      An AEAD primitive which uses the specified key.

    Raises:
      TinkError: If the key_uri is not supported.
    """
    if not self.does_support(key_uri):
      if self._key_uri:
        raise tink.TinkError(
            'This client is bound to %s and cannot use key %s' %
            (self._key_uri, key_uri))
      raise tink.TinkError(
          'This client does not support key %s' % key_uri)
    key_arn = _key_uri_to_key_arn(key_uri)
    session = boto3.session.Session(
        aws_access_key_id=self._aws_access_key_id,
        aws_secret_access_key=self._aws_secret_access_key,
        region_name=_get_region_from_key_arn(key_arn),
    )
    return _AwsKmsAead(session.client('kms'), key_arn)

  @classmethod
  def register_client(
      cls, key_uri: Optional[str], credentials_path: Optional[str]
  ) -> None:
    """Add a new KMS client to the global list of KMS clients.

    This function should only be called on startup and not on every operation.

    In many cases, it is not necessary to register the client. For example,
    you can create a KMS AEAD with
    kms_aead = awskms.AwsKmsClient(key_uri, credentials_path).get_aead(key_uri)
    and then use it to encrypt a keyset with KeysetHandle.write, or to create
    an envelope AEAD using aead.KmsEnvelopeAead.

    Args:
        key_uri: Optional key URI. If set, the registered client will only
          handle that key URI. If not set, then the client will handle all AWS
          KMS key URIs.
        credentials_path: Optional path to the credentials file. If it is not
          set, the default credentials are used.
    """
    tink.register_kms_client(AwsKmsClient(key_uri, credentials_path))



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for tink.python.tink.tink_config."""

import os
import tempfile

from absl.testing import absltest

import tink
from tink import aead
from tink import daead
from tink import hybrid
from tink import mac
from tink import prf
from tink import signature
from tink import streaming_aead
from tink import tink_config


def setUpModule():
  tink_config.register()


class TinkConfigTest(absltest.TestCase):

  def test_aead_encrypt_decrypt(self):
    keyset_handle = tink.new_keyset_handle(aead.aead_key_templates.AES256_GCM)
    primitive = keyset_handle.primitive(aead.Aead)
    self.assertEqual(
        primitive.decrypt(primitive.encrypt(b'plaintext', b'ad'), b'ad'),
        b'plaintext')

  def test_mac_compute_verify(self):
    keyset_handle = tink.new_keyset_handle(
        mac.mac_key_templates.HMAC_SHA256_128BITTAG)
    primitive = keyset_handle.primitive(mac.Mac)
    mac_value = primitive.compute_mac(b'data')
    self.assertIsNone(primitive.verify_mac(mac_value, b'data'))
    self.assertEqual(primitive.compute_mac(b'data'), mac_value)

  def test_deterministic_aead_wrapper_is_correctly_registered(self):
    keyset_handle = tink.new_keyset_handle(
        daead.deterministic_aead_key_templates.AES256_SIV)
    primitive = keyset_handle.primitive(daead.DeterministicAead)
    ciphertext = primitive.encrypt_deterministically(b'plaintext', b'ad')
    self.assertEqual(
        primitive.decrypt_deterministically(ciphertext, b'ad'), b'plaintext')
    self.assertEqual(
        primitive.encrypt_deterministically(b'plaintext', b'ad'), ciphertext)

  def test_hybrid_encrypt_decrypt(self):
    keyset_handle = tink.new_keyset_handle(
        hybrid.hybrid_key_templates.ECIES_P256_HKDF_HMAC_SHA256_AES128_GCM)
    decrypt = keyset_handle.primitive(hybrid.HybridDecrypt)
    public_keyset_handle = keyset_handle.public_keyset_handle()
    encrypt = public_keyset_handle.primitive(hybrid.HybridEncrypt)
    self.assertEqual(
        decrypt.decrypt(encrypt.encrypt(b'plaintext', b'ad'), b'ad'),
        b'plaintext')

  def test_signatures_sign_verify(self):
    keyset_handle = tink.new_keyset_handle(
        signature.signature_key_templates.ECDSA_P256)
    sign = keyset_handle.primitive(signature.PublicKeySign)
    public_keyset_handle = keyset_handle.public_keyset_handle()
    verify = public_keyset_handle.primitive(signature.PublicKeyVerify)
    sig = sign.sign(b'data')
    verify.verify(sig, b'data')

  def test_prf_compute(self):
    keyset_handle = tink.new_keyset_handle(prf.prf_key_templates.HMAC_SHA256)
    primitive = keyset_handle.primitive(prf.PrfSet)
    output = primitive.primary().compute(b'input_data', output_length=15)
    self.assertLen(output, 15)
    self.assertEqual(
        primitive.all()[primitive.primary_id()].compute(
            b'input_data', output_length=15), output)

  def test_streaming_aead_encrypt_decrypt(self):
    key_template = streaming_aead.streaming_aead_key_templates.AES128_GCM_HKDF_4KB
    keyset_handle = tink.new_keyset_handle(key_template)
    primitive = keyset_handle.primitive(streaming_aead.StreamingAead)
    plaintext = b'plaintext'
    associated_data = b'associated_data'
    with tempfile.TemporaryDirectory() as tmpdirname:
      filename = os.path.join(tmpdirname, 'encrypted_file')
      dest = open(filename, 'wb')
      with primitive.new_encrypting_stream(dest, associated_data) as es:
        es.write(plaintext)

      src = open(filename, 'rb')
      with primitive.new_decrypting_stream(src, associated_data) as ds:
        output = ds.read()
      self.assertEqual(output, plaintext)


if __name__ == '__main__':
  absltest.main()



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for tink.python.tink._keyset_handle."""

import io
# We are using Pickle only to test that Pickle doesn't work with keyset_handle.
import pickle  # pylint: disable=pickle-use

from absl.testing import absltest

from tink.proto import tink_pb2
import tink
from tink import aead
from tink import core
from tink import hybrid
from tink import mac
from tink import tink_config
from tink.testing import helper


def setUpModule():
  tink_config.register()


class FaultyAead(aead.Aead):

  def encrypt(self, plaintext: bytes, associated_data: bytes) -> bytes:
    raise core.TinkError('encrypt failed.')

  def decrypt(self, plaintext: bytes, associated_data: bytes) -> bytes:
    raise core.TinkError('decrypt failed.')


class BadAead1(aead.Aead):

  def encrypt(self, plaintext: bytes, associated_data: bytes) -> bytes:
    return b'ciphertext'

  def decrypt(self, plaintext: bytes, associated_data: bytes) -> bytes:
    return b'plaintext'


class BadAead2(aead.Aead):

  def encrypt(self, plaintext: bytes, associated_data: bytes) -> bytes:
    return b'ciphertext'

  def decrypt(self, plaintext: bytes, associated_data: bytes) -> bytes:
    return tink_pb2.Keyset(primary_key_id=42).SerializeToString()


def _master_key_aead():
  return core.Registry.primitive(
      core.Registry.new_key_data(aead.aead_key_templates.AES128_EAX), aead.Aead)


def _keyset_handle(keyset):
  # CleartextKeysetHandle is encouraged but we want to avoid that dependency
  # in the test
  return tink.KeysetHandle._create(keyset)


class KeysetHandleTest(absltest.TestCase):

  def test_instantiation__raises_error(self):
    with self.assertRaises(core.TinkError):
      tink.KeysetHandle()

  def test_generate_new(self):
    keyset_info = tink.new_keyset_handle(
        mac.mac_key_templates.HMAC_SHA256_128BITTAG).keyset_info()
    self.assertLen(keyset_info.key_info, 1)
    key_info = keyset_info.key_info[0]
    self.assertEqual(key_info.status, tink_pb2.ENABLED)
    self.assertEqual(
        key_info.output_prefix_type,
        mac.mac_key_templates.HMAC_SHA256_128BITTAG.output_prefix_type)
    self.assertEqual(keyset_info.primary_key_id, key_info.key_id)

  def test_generate_new_key_id_is_randomized(self):
    handle1 = tink.new_keyset_handle(
        mac.mac_key_templates.HMAC_SHA256_128BITTAG)
    handle2 = tink.new_keyset_handle(
        mac.mac_key_templates.HMAC_SHA256_128BITTAG)
    self.assertNotEqual(handle1.keyset_info().key_info[0].key_id,
                        handle2.keyset_info().key_info[0].key_id)

  def test_read_no_secret(self):
    private_handle = tink.new_keyset_handle(
        hybrid.hybrid_key_templates.ECIES_P256_HKDF_HMAC_SHA256_AES128_GCM)
    public_handle = private_handle.public_keyset_handle()

    output_stream_pub = io.BytesIO()
    writer = tink.BinaryKeysetWriter(output_stream_pub)
    writer.write(public_handle._keyset)

    output_stream_priv = io.BytesIO()
    writer = tink.BinaryKeysetWriter(output_stream_priv)
    writer.write(private_handle._keyset)

    reader = tink.BinaryKeysetReader(output_stream_pub.getvalue())
    tink.read_no_secret_keyset_handle(reader)

    reader = tink.BinaryKeysetReader(output_stream_priv.getvalue())
    with self.assertRaises(core.TinkError):
      tink.read_no_secret_keyset_handle(reader)

  def test_write_no_secret(self):
    private_handle = tink.new_keyset_handle(
        hybrid.hybrid_key_templates.ECIES_P256_HKDF_HMAC_SHA256_AES128_GCM)
    public_handle = private_handle.public_keyset_handle()

    output_stream = io.BytesIO()
    writer = tink.BinaryKeysetWriter(output_stream)

    public_handle.write_no_secret(writer)

    with self.assertRaises(core.TinkError):
      private_handle.write_no_secret(writer)

  def test_write_encrypted(self):
    handle = tink.new_keyset_handle(mac.mac_key_templates.HMAC_SHA256_128BITTAG)
    # Encrypt the keyset with Aead.
    master_key_aead = _master_key_aead()
    output_stream = io.BytesIO()
    writer = tink.BinaryKeysetWriter(output_stream)
    handle.write(writer, master_key_aead)
    reader = tink.BinaryKeysetReader(output_stream.getvalue())
    handle2 = tink.read_keyset_handle(reader, master_key_aead)
    # Check that handle2 has the same primitive as handle.
    handle2.primitive(mac.Mac).verify_mac(
        handle.primitive(mac.Mac).compute_mac(b'data'), b'data')

  def test_write_encrypted_with_associated_data(self):
    handle = tink.new_keyset_handle(mac.mac_key_templates.HMAC_SHA256_128BITTAG)
    # Encrypt the keyset with Aead.
    master_key_aead = _master_key_aead()
    output_stream = io.BytesIO()
    writer = tink.BinaryKeysetWriter(output_stream)
    handle.write_with_associated_data(writer, master_key_aead, b'01')
    reader = tink.BinaryKeysetReader(output_stream.getvalue())
    handle2 = tink.read_keyset_handle_with_associated_data(
        reader, master_key_aead, b'01')
    # Check that handle2 has the same primitive as handle.
    handle2.primitive(mac.Mac).verify_mac(
        handle.primitive(mac.Mac).compute_mac(b'data'), b'data')

  def test_write_encrypted_with_mismatched_associated_data(self):
    handle = tink.new_keyset_handle(mac.mac_key_templates.HMAC_SHA256_128BITTAG)
    # Encrypt the keyset with Aead.
    master_key_aead = _master_key_aead()
    output_stream = io.BytesIO()
    writer = tink.BinaryKeysetWriter(output_stream)
    handle.write_with_associated_data(writer, master_key_aead, b'01')
    reader = tink.BinaryKeysetReader(output_stream.getvalue())
    with self.assertRaises(core.TinkError):
      tink.read_keyset_handle_with_associated_data(reader, master_key_aead,
                                                   b'02')

  def test_write_raises_error_when_encrypt_failed(self):
    handle = tink.new_keyset_handle(mac.mac_key_templates.HMAC_SHA256_128BITTAG)
    writer = tink.BinaryKeysetWriter(io.BytesIO())
    with self.assertRaises(core.TinkError):
      handle.write(writer, FaultyAead())

  # Not clear what this tests.
  def test_write_raises_error_when_decrypt_not_possible(self):
    handle = tink.new_keyset_handle(mac.mac_key_templates.HMAC_SHA256_128BITTAG)
    writer = tink.BinaryKeysetWriter(io.BytesIO())
    with self.assertRaises(core.TinkError):
      handle.write(writer, BadAead1())

  def test_write_raises_error_when_decrypt_to_wrong_keyset(self):
    handle = tink.new_keyset_handle(mac.mac_key_templates.HMAC_SHA256_128BITTAG)
    writer = tink.BinaryKeysetWriter(io.BytesIO())
    with self.assertRaises(core.TinkError):
      handle.write(writer, BadAead2())

  def test_read_empty_keyset_fails(self):
    with self.assertRaises(core.TinkError):
      tink.read_keyset_handle(tink.BinaryKeysetReader(b''), _master_key_aead())

  def test_public_keyset_handle(self):
    private_handle = tink.new_keyset_handle(
        hybrid.hybrid_key_templates.ECIES_P256_HKDF_HMAC_SHA256_AES128_GCM)
    public_handle = private_handle.public_keyset_handle()
    hybrid_dec = private_handle.primitive(hybrid.HybridDecrypt)
    hybrid_enc = public_handle.primitive(hybrid.HybridEncrypt)

    self.assertEqual(public_handle.keyset_info().primary_key_id,
                     private_handle.keyset_info().primary_key_id)
    self.assertLen(public_handle.keyset_info().key_info, 1)
    self.assertEqual(
        public_handle.keyset_info().key_info[0].type_url,
        'type.googleapis.com/google.crypto.tink.EciesAeadHkdfPublicKey')

    ciphertext = hybrid_enc.encrypt(b'some plaintext', b'some context info')
    self.assertEqual(
        hybrid_dec.decrypt(ciphertext, b'some context info'), b'some plaintext')

  def test_primitive_success(self):
    keyset = tink_pb2.Keyset()
    key = keyset.key.add()
    key.key_data.CopyFrom(
        core.Registry.new_key_data(aead.aead_key_templates.AES128_EAX))
    key.output_prefix_type = tink_pb2.TINK
    key.key_id = 1
    key.status = tink_pb2.ENABLED
    keyset.primary_key_id = 1
    handle = _keyset_handle(keyset)
    aead_primitive = handle.primitive(aead.Aead)
    self.assertEqual(
        aead_primitive.decrypt(
            aead_primitive.encrypt(b'message', b'aad'), b'aad'), b'message')

  def test_init_with_fake_key_success(self):
    keyset = tink_pb2.Keyset()
    keyset.key.extend([helper.fake_key(key_id=123)])
    keyset.primary_key_id = 123
    _ = _keyset_handle(keyset)

  def test_init_fails_on_keyset_with_only_destroyed_keys(self):
    keyset = tink_pb2.Keyset()
    keyset.key.extend([helper.fake_key(key_id=123, status=tink_pb2.DESTROYED)])
    keyset.primary_key_id = 123
    with self.assertRaises(core.TinkError):
      _ = _keyset_handle(keyset)

  def test_init_fails_on_key_without_keydata(self):
    keyset = tink_pb2.Keyset()
    key = helper.fake_key(key_id=123)
    key.ClearField('key_data')
    keyset.key.extend([key])
    keyset.primary_key_id = 123
    with self.assertRaises(core.TinkError):
      handle = _keyset_handle(keyset)
      handle.primitive(aead.Aead)

  def test_init_fails_on_key_with_unknown_prefix(self):
    keyset = tink_pb2.Keyset()
    keyset.key.extend([
        helper.fake_key(key_id=123, output_prefix_type=tink_pb2.UNKNOWN_PREFIX)
    ])
    keyset.primary_key_id = 123
    with self.assertRaises(core.TinkError):
      _ = _keyset_handle(keyset)

  def test_init_fails_on_key_with_unknown_status(self):
    keyset = tink_pb2.Keyset()
    keyset.key.extend(
        [helper.fake_key(key_id=123, status=tink_pb2.UNKNOWN_STATUS)])
    keyset.primary_key_id = 123
    with self.assertRaises(core.TinkError):
      _ = _keyset_handle(keyset)

  def test_init_fails_on_multiple_primary_keys(self):
    keyset = tink_pb2.Keyset()
    keyset.key.extend(
        [helper.fake_key(key_id=123),
         helper.fake_key(key_id=123)])
    keyset.primary_key_id = 123
    with self.assertRaises(core.TinkError):
      _ = _keyset_handle(keyset)

  def test_init_fails_without_primary_key_present(self):
    keyset = tink_pb2.Keyset()
    keyset.key.extend([helper.fake_key(key_id=123)])
    with self.assertRaises(core.TinkError):
      _ = _keyset_handle(keyset)

  def test_primitive_fails_on_wrong_primitive_class(self):
    keyset = tink_pb2.Keyset()
    key = keyset.key.add()
    key.key_data.CopyFrom(
        core.Registry.new_key_data(aead.aead_key_templates.AES128_EAX))
    key.output_prefix_type = tink_pb2.TINK
    key.key_id = 1
    key.status = tink_pb2.ENABLED
    keyset.primary_key_id = 1
    handle = _keyset_handle(keyset)
    _ = handle.primitive(aead.Aead)  # Make sure the handle works for AEAD.
    with self.assertRaises(core.TinkError):
      handle.primitive(mac.Mac)

  def test_primitive_wrapped_correctly(self):
    keydata2 = core.Registry.new_key_data(aead.aead_key_templates.AES128_EAX)
    keyset = tink_pb2.Keyset()
    key = keyset.key.add()
    key.key_data.CopyFrom(
        core.Registry.new_key_data(aead.aead_key_templates.AES128_EAX))
    key.output_prefix_type = tink_pb2.TINK
    key.key_id = 1
    key.status = tink_pb2.ENABLED
    key = keyset.key.add()
    key.key_data.CopyFrom(keydata2)
    key.output_prefix_type = tink_pb2.RAW
    key.key_id = 2
    key.status = tink_pb2.ENABLED
    keyset.primary_key_id = 1
    handle = _keyset_handle(keyset)
    aead_primitive = handle.primitive(aead.Aead)
    aead_primitive2 = core.Registry.primitive(keydata2, aead.Aead)
    self.assertEqual(
        aead_primitive.decrypt(
            aead_primitive2.encrypt(b'message', b'aad'), b'aad'), b'message')

  def test_keyset_info(self):
    keyset = tink_pb2.Keyset(primary_key_id=1)
    keyset.key.extend([
        helper.fake_key(
            value=b'v1',
            type_url='t1',
            key_id=1,
            status=tink_pb2.ENABLED,
            output_prefix_type=tink_pb2.TINK),
        helper.fake_key(
            value=b'v2',
            type_url='t2',
            key_id=2,
            status=tink_pb2.DESTROYED,
            output_prefix_type=tink_pb2.RAW)
    ])
    handle = _keyset_handle(keyset)
    expected_keyset_info = tink_pb2.KeysetInfo(primary_key_id=1)
    info1 = expected_keyset_info.key_info.add()
    info1.type_url = 't1'
    info1.status = tink_pb2.ENABLED
    info1.output_prefix_type = tink_pb2.TINK
    info1.key_id = 1
    info2 = expected_keyset_info.key_info.add()
    info2.type_url = 't2'
    info2.status = tink_pb2.DESTROYED
    info2.output_prefix_type = tink_pb2.RAW
    info2.key_id = 2
    self.assertEqual(expected_keyset_info, handle.keyset_info())

  def test_pickle_dumps_fails(self):
    handle = tink.new_keyset_handle(mac.mac_key_templates.HMAC_SHA256_128BITTAG)
    with self.assertRaises(tink.TinkError):
      _ = pickle.dumps(handle)

  def test_key_template_with_invalid_output_prefix_type_fails(self):
    template = aead.aead_key_templates.XAES_256_GCM_160_BIT_NONCE_NO_PREFIX
    template.output_prefix_type = tink_pb2.LEGACY
    with self.assertRaises(tink.TinkError):
      tink.new_keyset_handle(template)


if __name__ == '__main__':
  absltest.main()



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Interface for PublicKeySign."""

import abc


class PublicKeySign(metaclass=abc.ABCMeta):
  """Interface for public key signing.

  Digital Signatures provide functionality of signing data and verification of
  the signatures. They are represented by a pair of primitives (interfaces)
  'PublicKeySign' for signing of data, and 'PublicKeyVerify' for verification
  of signatures. Implementations of these interfaces are secure against
  adaptive chosen-message attacks. Signing data ensures the authenticity and
  the integrity of that data, but not its secrecy.
  """

  @abc.abstractmethod
  def sign(self, data: bytes) -> bytes:
    """Computes the signature for data.

    Args:
      data: bytes, the input data.
    Returns:
      The signature as bytes.
    """
    raise NotImplementedError()



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Python wrapper of the wrapped C++ MAC key manager."""

from tink import core
from tink.cc.pybind import tink_bindings
from tink.mac import _mac
from tink.mac import _mac_wrapper


class _MacCcToPyWrapper(_mac.Mac):
  """Transforms C++ Mac primitive into a Python primitive."""

  def __init__(self, cc_primitive: tink_bindings.Mac):
    self._cc_mac = cc_primitive

  @core.use_tink_errors
  def compute_mac(self, data: bytes) -> bytes:
    return self._cc_mac.compute_mac(data)

  @core.use_tink_errors
  def verify_mac(self, mac_value: bytes, data: bytes) -> None:
    self._cc_mac.verify_mac(mac_value, data)


def register():
  tink_bindings.register()

  for key_type_identifier in ('HmacKey', 'AesCmacKey',):
    type_url = 'type.googleapis.com/google.crypto.tink.' + key_type_identifier
    key_manager = core.KeyManagerCcToPyWrapper(
        tink_bindings.MacKeyManager.from_cc_registry(type_url), _mac.Mac,
        _MacCcToPyWrapper)
    core.Registry.register_key_manager(key_manager, new_key_allowed=True)
  core.Registry.register_primitive_wrapper(_mac_wrapper.MacWrapper())



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Python wrapper of the wrapped C++ AEAD key manager."""

from tink import core
from tink.aead import _aead
from tink.aead import _aead_wrapper
from tink.aead import _kms_aead_key_manager
from tink.cc.pybind import tink_bindings


class AeadCcToPyWrapper(_aead.Aead):
  """Transforms C++ Aead primitive into a Python primitive."""

  def __init__(self, cc_primitive: tink_bindings.Aead):
    self._aead = cc_primitive

  @core.use_tink_errors
  def encrypt(self, plaintext: bytes, associated_data: bytes) -> bytes:
    return self._aead.encrypt(plaintext, associated_data)

  @core.use_tink_errors
  def decrypt(self, plaintext: bytes, associated_data: bytes) -> bytes:
    return self._aead.decrypt(plaintext, associated_data)


def register() -> None:
  """Registers all AEAD key managers and AEAD wrapper in the Registry."""
  tink_bindings.register()
  for ident in (
      'AesCtrHmacAeadKey',
      'AesGcmKey',
      'AesGcmSivKey',
      'AesEaxKey',
      'XChaCha20Poly1305Key',
      'XAesGcmKey',
  ):
    type_url = 'type.googleapis.com/google.crypto.tink.{}'.format(ident)
    key_manager = core.KeyManagerCcToPyWrapper(
        tink_bindings.AeadKeyManager.from_cc_registry(type_url), _aead.Aead,
        AeadCcToPyWrapper)
    core.Registry.register_key_manager(key_manager, new_key_allowed=True)
  core.Registry.register_primitive_wrapper(_aead_wrapper.AeadWrapper())
  core.Registry.register_key_manager(
      _kms_aead_key_manager.KmsAeadKeyManager(), new_key_allowed=True
  )
  core.Registry.register_key_manager(
      _kms_aead_key_manager.KmsEnvelopeAeadKeyManager(), new_key_allowed=True
  )



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# NO CHECKED-IN PROTOBUF GENCODE
# source: tink/proto/slh_dsa.proto
# Protobuf Python Version: 6.30.2
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import runtime_version as _runtime_version
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
_runtime_version.ValidateProtobufRuntimeVersion(
    _runtime_version.Domain.PUBLIC,
    6,
    30,
    2,
    '',
    'tink/proto/slh_dsa.proto'
)
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x18tink/proto/slh_dsa.proto\x12\x12google.crypto.tink\"\x92\x01\n\x0cSlhDsaParams\x12\x10\n\x08key_size\x18\x01 \x01(\x05\x12\x35\n\thash_type\x18\x02 \x01(\x0e\x32\".google.crypto.tink.SlhDsaHashType\x12\x39\n\x08sig_type\x18\x03 \x01(\x0e\x32\'.google.crypto.tink.SlhDsaSignatureType\"T\n\x0fSlhDsaKeyFormat\x12\x0f\n\x07version\x18\x01 \x01(\r\x12\x30\n\x06params\x18\x02 \x01(\x0b\x32 .google.crypto.tink.SlhDsaParams\"g\n\x0fSlhDsaPublicKey\x12\x0f\n\x07version\x18\x01 \x01(\r\x12\x11\n\tkey_value\x18\x02 \x01(\x0c\x12\x30\n\x06params\x18\x03 \x01(\x0b\x32 .google.crypto.tink.SlhDsaParams\"o\n\x10SlhDsaPrivateKey\x12\x0f\n\x07version\x18\x01 \x01(\r\x12\x11\n\tkey_value\x18\x02 \x01(\x0c\x12\x37\n\npublic_key\x18\x03 \x01(\x0b\x32#.google.crypto.tink.SlhDsaPublicKey*H\n\x0eSlhDsaHashType\x12!\n\x1dSLH_DSA_HASH_TYPE_UNSPECIFIED\x10\x00\x12\x08\n\x04SHA2\x10\x01\x12\t\n\x05SHAKE\x10\x02*d\n\x13SlhDsaSignatureType\x12&\n\"SLH_DSA_SIGNATURE_TYPE_UNSPECIFIED\x10\x00\x12\x10\n\x0c\x46\x41ST_SIGNING\x10\x01\x12\x13\n\x0fSMALL_SIGNATURE\x10\x02\x42W\n\x1c\x63om.google.crypto.tink.protoP\x01Z5github.com/tink-crypto/tink-go/v2/proto/slh_dsa_protob\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'tink.proto.slh_dsa_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  _globals['DESCRIPTOR']._loaded_options = None
  _globals['DESCRIPTOR']._serialized_options = b'\n\034com.google.crypto.tink.protoP\001Z5github.com/tink-crypto/tink-go/v2/proto/slh_dsa_proto'
  _globals['_SLHDSAHASHTYPE']._serialized_start=501
  _globals['_SLHDSAHASHTYPE']._serialized_end=573
  _globals['_SLHDSASIGNATURETYPE']._serialized_start=575
  _globals['_SLHDSASIGNATURETYPE']._serialized_end=675
  _globals['_SLHDSAPARAMS']._serialized_start=49
  _globals['_SLHDSAPARAMS']._serialized_end=195
  _globals['_SLHDSAKEYFORMAT']._serialized_start=197
  _globals['_SLHDSAKEYFORMAT']._serialized_end=281
  _globals['_SLHDSAPUBLICKEY']._serialized_start=283
  _globals['_SLHDSAPUBLICKEY']._serialized_end=386
  _globals['_SLHDSAPRIVATEKEY']._serialized_start=388
  _globals['_SLHDSAPRIVATEKEY']._serialized_end=499
# @@protoc_insertion_point(module_scope)



# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
"""Pre-generated JWT KeyTemplate."""

from tink.proto import jwt_ecdsa_pb2
from tink.proto import jwt_hmac_pb2
from tink.proto import jwt_rsa_ssa_pkcs1_pb2
from tink.proto import jwt_rsa_ssa_pss_pb2
from tink.proto import tink_pb2
from tink.internal import big_integer_util


_F4 = 65537


def _create_jwt_hmac_template(
    algorithm: jwt_hmac_pb2.JwtHmacAlgorithm, key_size: int,
    output_prefix_type: tink_pb2.OutputPrefixType) -> tink_pb2.KeyTemplate:
  key_format = jwt_hmac_pb2.JwtHmacKeyFormat(
      algorithm=algorithm, key_size=key_size)
  return tink_pb2.KeyTemplate(
      type_url='type.googleapis.com/google.crypto.tink.JwtHmacKey',
      value=key_format.SerializeToString(),
      output_prefix_type=output_prefix_type)


def _create_jwt_ecdsa_template(
    algorithm: jwt_ecdsa_pb2.JwtEcdsaAlgorithm,
    output_prefix_type: tink_pb2.OutputPrefixType) -> tink_pb2.KeyTemplate:
  key_format = jwt_ecdsa_pb2.JwtEcdsaKeyFormat(
      algorithm=algorithm)
  return tink_pb2.KeyTemplate(
      type_url='type.googleapis.com/google.crypto.tink.JwtEcdsaPrivateKey',
      value=key_format.SerializeToString(),
      output_prefix_type=output_prefix_type)


def _create_jwt_rsa_ssa_pkcs1_template(
    algorithm: jwt_rsa_ssa_pkcs1_pb2.JwtRsaSsaPkcs1Algorithm, modulus_size: int,
    output_prefix_type: tink_pb2.OutputPrefixType) -> tink_pb2.KeyTemplate:
  key_format = jwt_rsa_ssa_pkcs1_pb2.JwtRsaSsaPkcs1KeyFormat(
      algorithm=algorithm,
      modulus_size_in_bits=modulus_size,
      public_exponent=big_integer_util.num_to_bytes(_F4))
  return tink_pb2.KeyTemplate(
      type_url='type.googleapis.com/google.crypto.tink.JwtRsaSsaPkcs1PrivateKey',
      value=key_format.SerializeToString(),
      output_prefix_type=output_prefix_type)


def _create_jwt_rsa_ssa_pss_template(
    algorithm: jwt_rsa_ssa_pss_pb2.JwtRsaSsaPssAlgorithm, modulus_size: int,
    output_prefix_type: tink_pb2.OutputPrefixType) -> tink_pb2.KeyTemplate:
  key_format = jwt_rsa_ssa_pss_pb2.JwtRsaSsaPssKeyFormat(
      algorithm=algorithm,
      modulus_size_in_bits=modulus_size,
      public_exponent=big_integer_util.num_to_bytes(_F4))
  return tink_pb2.KeyTemplate(
      type_url='type.googleapis.com/google.crypto.tink.JwtRsaSsaPssPrivateKey',
      value=key_format.SerializeToString(),
      output_prefix_type=output_prefix_type)


# Hmac Templates
def jwt_hs256_template() -> tink_pb2.KeyTemplate:
  return _create_jwt_hmac_template(jwt_hmac_pb2.HS256, 32, tink_pb2.TINK)


def raw_jwt_hs256_template() -> tink_pb2.KeyTemplate:
  return _create_jwt_hmac_template(jwt_hmac_pb2.HS256, 32, tink_pb2.RAW)


def jwt_hs384_template() -> tink_pb2.KeyTemplate:
  return _create_jwt_hmac_template(jwt_hmac_pb2.HS384, 48, tink_pb2.TINK)


def raw_jwt_hs384_template() -> tink_pb2.KeyTemplate:
  return _create_jwt_hmac_template(jwt_hmac_pb2.HS384, 48, tink_pb2.RAW)


def jwt_hs512_template() -> tink_pb2.KeyTemplate:
  return _create_jwt_hmac_template(jwt_hmac_pb2.HS512, 64, tink_pb2.TINK)


def raw_jwt_hs512_template() -> tink_pb2.KeyTemplate:
  return _create_jwt_hmac_template(jwt_hmac_pb2.HS512, 64, tink_pb2.RAW)


# ECDSA Templates
def jwt_es256_template() -> tink_pb2.KeyTemplate:
  return _create_jwt_ecdsa_template(jwt_ecdsa_pb2.ES256, tink_pb2.TINK)


def raw_jwt_es256_template() -> tink_pb2.KeyTemplate:
  return _create_jwt_ecdsa_template(jwt_ecdsa_pb2.ES256, tink_pb2.RAW)


def jwt_es384_template() -> tink_pb2.KeyTemplate:
  return _create_jwt_ecdsa_template(jwt_ecdsa_pb2.ES384, tink_pb2.TINK)


def raw_jwt_es384_template() -> tink_pb2.KeyTemplate:
  return _create_jwt_ecdsa_template(jwt_ecdsa_pb2.ES384, tink_pb2.RAW)


def jwt_es512_template() -> tink_pb2.KeyTemplate:
  return _create_jwt_ecdsa_template(jwt_ecdsa_pb2.ES512, tink_pb2.TINK)


def raw_jwt_es512_template() -> tink_pb2.KeyTemplate:
  return _create_jwt_ecdsa_template(jwt_ecdsa_pb2.ES512, tink_pb2.RAW)


# RSA SSA PKCS1 Templates
def jwt_rs256_2048_f4_template() -> tink_pb2.KeyTemplate:
  return _create_jwt_rsa_ssa_pkcs1_template(jwt_rsa_ssa_pkcs1_pb2.RS256, 2048,
                                            tink_pb2.TINK)


def raw_jwt_rs256_2048_f4_template() -> tink_pb2.KeyTemplate:
  return _create_jwt_rsa_ssa_pkcs1_template(jwt_rsa_ssa_pkcs1_pb2.RS256, 2048,
                                            tink_pb2.RAW)


def jwt_rs256_3072_f4_template() -> tink_pb2.KeyTemplate:
  return _create_jwt_rsa_ssa_pkcs1_template(jwt_rsa_ssa_pkcs1_pb2.RS256, 3072,
                                            tink_pb2.TINK)


def raw_jwt_rs256_3072_f4_template() -> tink_pb2.KeyTemplate:
  return _create_jwt_rsa_ssa_pkcs1_template(jwt_rsa_ssa_pkcs1_pb2.RS256, 3072,
                                            tink_pb2.RAW)


def jwt_rs384_3072_f4_template() -> tink_pb2.KeyTemplate:
  return _create_jwt_rsa_ssa_pkcs1_template(jwt_rsa_ssa_pkcs1_pb2.RS384, 3072,
                                            tink_pb2.TINK)


def raw_jwt_rs384_3072_f4_template() -> tink_pb2.KeyTemplate:
  return _create_jwt_rsa_ssa_pkcs1_template(jwt_rsa_ssa_pkcs1_pb2.RS384, 3072,
                                            tink_pb2.RAW)


def jwt_rs512_4096_f4_template() -> tink_pb2.KeyTemplate:
  return _create_jwt_rsa_ssa_pkcs1_template(jwt_rsa_ssa_pkcs1_pb2.RS512, 4096,
                                            tink_pb2.TINK)


def raw_jwt_rs512_4096_f4_template() -> tink_pb2.KeyTemplate:
  return _create_jwt_rsa_ssa_pkcs1_template(jwt_rsa_ssa_pkcs1_pb2.RS512, 4096,
                                            tink_pb2.RAW)


# RSA SSA PSS Templates
def jwt_ps256_2048_f4_template() -> tink_pb2.KeyTemplate:
  return _create_jwt_rsa_ssa_pss_template(jwt_rsa_ssa_pss_pb2.PS256, 2048,
                                          tink_pb2.TINK)


def raw_jwt_ps256_2048_f4_template() -> tink_pb2.KeyTemplate:
  return _create_jwt_rsa_ssa_pss_template(jwt_rsa_ssa_pss_pb2.PS256, 2048,
                                          tink_pb2.RAW)


def jwt_ps256_3072_f4_template() -> tink_pb2.KeyTemplate:
  return _create_jwt_rsa_ssa_pss_template(jwt_rsa_ssa_pss_pb2.PS256, 3072,
                                          tink_pb2.TINK)


def raw_jwt_ps256_3072_f4_template() -> tink_pb2.KeyTemplate:
  return _create_jwt_rsa_ssa_pss_template(jwt_rsa_ssa_pss_pb2.PS256, 3072,
                                          tink_pb2.RAW)


def jwt_ps384_3072_f4_template() -> tink_pb2.KeyTemplate:
  return _create_jwt_rsa_ssa_pss_template(jwt_rsa_ssa_pss_pb2.PS384, 3072,
                                          tink_pb2.TINK)


def raw_jwt_ps384_3072_f4_template() -> tink_pb2.KeyTemplate:
  return _create_jwt_rsa_ssa_pss_template(jwt_rsa_ssa_pss_pb2.PS384, 3072,
                                          tink_pb2.RAW)


def jwt_ps512_4096_f4_template() -> tink_pb2.KeyTemplate:
  return _create_jwt_rsa_ssa_pss_template(jwt_rsa_ssa_pss_pb2.PS512, 4096,
                                          tink_pb2.TINK)


def raw_jwt_ps512_4096_f4_template() -> tink_pb2.KeyTemplate:
  return _create_jwt_rsa_ssa_pss_template(jwt_rsa_ssa_pss_pb2.PS512, 4096,
                                          tink_pb2.RAW)



# Copyright 2023 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Integration tests for Tink Python's HashiCorp Vault KMS integration."""

import base64
import os

from absl.testing import absltest
import hvac

import tink
from tink import aead
from tink.integration import hcvault


_VAULT_TOKEN = os.getenv('VAULT_TOKEN', '')  # Your auth token
_VAULT_ADDR = os.getenv('VAULT_ADDR', '')

_BAD_TOKEN = 'notavalidtoken'

# Replace this with your vault URI
_KEY_PATH = 'transit/keys/key-1'


def setUpModule():
  aead.register()


def _corrupt(value: bytes, i: int) -> bytes:
  """Corrupts a byte in `value` at a given index `i`."""
  assert i >= 0 and i < len(value)
  tmp_value = list(value)
  tmp_value[i] ^= 2
  return bytes(tmp_value)


class HcVaultAeadTest(absltest.TestCase):

  def setUp(self):
    super().setUp()
    self.client = hvac.Client(url=_VAULT_ADDR, token=_VAULT_TOKEN, verify=False)

  def test_encrypt_decrypt(self):
    vaultaead = hcvault.new_aead(_KEY_PATH, self.client)

    plaintext = bytes(i for i in range(256))
    ciphertext = vaultaead.encrypt(plaintext, associated_data=b'')
    self.assertEqual(
        plaintext, vaultaead.decrypt(ciphertext, associated_data=b'')
    )

  def test_corrupted_ciphertext(self):
    vaultaead = hcvault.new_aead(_KEY_PATH, self.client)

    plaintext = b'helloworld'
    ciphertext = vaultaead.encrypt(plaintext, b'')
    self.assertEqual(plaintext, vaultaead.decrypt(ciphertext, b''))

    # The returned ciphertext is of the form:
    #           vault:v{N}:Base64(IV+Ciphertext)
    vault, version, iv_and_ciphertext = ciphertext.decode().split(':')
    # Corrupt vault.
    for i in range(len(vault)):
      corrupted_vault = _corrupt(vault.encode(), i)
      with self.assertRaises(tink.TinkError):
        vaultaead.decrypt(
            f'{corrupted_vault}:{version}:{iv_and_ciphertext}'.encode(), b''
        )

    # Corrupt the version.
    for i in range(len(version)):
      corrupted_version = _corrupt(version.encode(), i)
      with self.assertRaises(tink.TinkError):
        vaultaead.decrypt(
            f'{vault}:{corrupted_version}:{iv_and_ciphertext}'.encode(), b''
        )

    # Corrupt the ciphertext.
    # In this case we corrupt the decoded string, then encode back to Base64.
    iv_and_ciphertext = base64.b64decode(iv_and_ciphertext.encode())
    for i in range(len(iv_and_ciphertext)):
      corrupted_iv_and_ciphertext = base64.b64encode(
          _corrupt(iv_and_ciphertext, i)
      )
      with self.assertRaises(tink.TinkError):
        vaultaead.decrypt(
            f'{vault}:{version}:{corrupted_iv_and_ciphertext}', b''
        )

  def test_encrypt_with_bad_client(self):
    bad_client = hvac.Client(url=_VAULT_ADDR, token=_BAD_TOKEN, verify=False)
    vaultaead = hcvault.new_aead(_KEY_PATH, bad_client)
    with self.assertRaises(tink.TinkError):
      vaultaead.encrypt(plaintext=b'hello', associated_data=b'')


if __name__ == '__main__':
  absltest.main()



# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""JWT Signature key managers."""

from typing import Any, Optional, Type, Tuple, Callable

from tink.proto import jwt_ecdsa_pb2
from tink.proto import jwt_rsa_ssa_pkcs1_pb2
from tink.proto import jwt_rsa_ssa_pss_pb2
from tink.proto import tink_pb2
from tink import core
from tink.cc.pybind import tink_bindings
from tink.jwt import _json_util
from tink.jwt import _jwt_error
from tink.jwt import _jwt_format
from tink.jwt import _jwt_public_key_sign
from tink.jwt import _jwt_public_key_verify
from tink.jwt import _jwt_validator
from tink.jwt import _raw_jwt
from tink.jwt import _verified_jwt

_JWT_ECDSA_PRIVATE_KEY_TYPE = 'type.googleapis.com/google.crypto.tink.JwtEcdsaPrivateKey'
_JWT_ECDSA_PUBLIC_KEY_TYPE = 'type.googleapis.com/google.crypto.tink.JwtEcdsaPublicKey'

_JWT_RSA_SSA_PKCS1_PRIVATE_KEY_TYPE = 'type.googleapis.com/google.crypto.tink.JwtRsaSsaPkcs1PrivateKey'
_JWT_RSA_SSA_PKCS1_PUBLIC_KEY_TYPE = 'type.googleapis.com/google.crypto.tink.JwtRsaSsaPkcs1PublicKey'

_JWT_RSA_SSA_PSS_PRIVATE_KEY_TYPE = 'type.googleapis.com/google.crypto.tink.JwtRsaSsaPssPrivateKey'
_JWT_RSA_SSA_PSS_PUBLIC_KEY_TYPE = 'type.googleapis.com/google.crypto.tink.JwtRsaSsaPssPublicKey'

_ECDSA_ALGORITHM_TEXTS = {
    jwt_ecdsa_pb2.ES256: 'ES256',
    jwt_ecdsa_pb2.ES384: 'ES384',
    jwt_ecdsa_pb2.ES512: 'ES512'
}

_RSA_SSA_PKCS1_ALGORITHM_TEXTS = {
    jwt_rsa_ssa_pkcs1_pb2.RS256: 'RS256',
    jwt_rsa_ssa_pkcs1_pb2.RS384: 'RS384',
    jwt_rsa_ssa_pkcs1_pb2.RS512: 'RS512'
}

_RSA_SSA_PSS_ALGORITHM_TEXTS = {
    jwt_rsa_ssa_pss_pb2.PS256: 'PS256',
    jwt_rsa_ssa_pss_pb2.PS384: 'PS384',
    jwt_rsa_ssa_pss_pb2.PS512: 'PS512'
}


class _JwtPublicKeySign(_jwt_public_key_sign.JwtPublicKeySignInternal):
  """Implementation of JwtPublicKeySignInternal using a PublicKeySign."""

  def __init__(self, cc_primitive: tink_bindings.PublicKeySign, algorithm: str,
               custom_kid: Optional[str]):
    self._public_key_sign = cc_primitive
    self._algorithm = algorithm
    self._custom_kid = custom_kid

  @core.use_tink_errors
  def _sign(self, data: bytes) -> bytes:
    return self._public_key_sign.sign(data)

  def sign_and_encode_with_kid(self, raw_jwt: _raw_jwt.RawJwt,
                               kid: Optional[str]) -> str:
    """Computes a signature and encodes the token.

    Args:
      raw_jwt: The RawJwt token to be MACed and encoded.
      kid: Optional "kid" header value. It is set by the wrapper for keys with
        output prefix TINK, and it is None for output prefix RAW.

    Returns:
      The MACed token encoded in the JWS compact serialization format.
    Raises:
      tink.TinkError if the operation fails.
    """
    if self._custom_kid is not None:
      if kid is not None:
        raise _jwt_error.JwtInvalidError(
            'custom_kid must not be set for keys with output prefix type TINK')
      kid = self._custom_kid
    unsigned = _jwt_format.create_unsigned_compact(self._algorithm, kid,
                                                   raw_jwt)
    return _jwt_format.create_signed_compact(unsigned, self._sign(unsigned))


class _JwtPublicKeyVerify(_jwt_public_key_verify.JwtPublicKeyVerifyInternal):
  """Implementation of JwtPublicKeyVerify using a PublicKeyVerify."""

  def __init__(self, cc_primitive: tink_bindings.PublicKeyVerify,
               algorithm: str, custom_kid: Optional[str]):
    self._public_key_verify = cc_primitive
    self._algorithm = algorithm
    self._custom_kid = custom_kid

  @core.use_tink_errors
  def _verify(self, signature: bytes, data: bytes) -> None:
    self._public_key_verify.verify(signature, data)

  def verify_and_decode_with_kid(
      self, compact: str, validator: _jwt_validator.JwtValidator,
      kid: Optional[str]) -> _verified_jwt.VerifiedJwt:
    """Verifies, validates and decodes a signed compact JWT token."""
    parts = _jwt_format.split_signed_compact(compact)
    unsigned_compact, json_header, json_payload, signature = parts
    self._verify(signature, unsigned_compact)
    header = _json_util.json_loads(json_header)
    _jwt_format.validate_header(
        header=header,
        algorithm=self._algorithm,
        tink_kid=kid,
        custom_kid=self._custom_kid)
    raw_jwt = _raw_jwt.raw_jwt_from_json(
        _jwt_format.get_type_header(header), json_payload)
    _jwt_validator.validate(validator, raw_jwt)
    return _verified_jwt.VerifiedJwt._create(raw_jwt)  # pylint: disable=protected-access


class _JwtPublicKeySignKeyManagerCcToPyWrapper(
    core.PrivateKeyManager[_jwt_public_key_sign.JwtPublicKeySignInternal]):
  """Converts a C++ sign key manager into a JwtPublicKeySignKeyManager."""

  def __init__(self, cc_key_manager: tink_bindings.PublicKeySignKeyManager,
               key_data_to_alg_kid: Callable[[tink_pb2.KeyData],
                                             Tuple[str, Optional[str]]]):
    self._cc_key_manager = cc_key_manager
    self._key_data_to_alg_kid = key_data_to_alg_kid

  def primitive_class(
      self) -> Type[_jwt_public_key_sign.JwtPublicKeySignInternal]:
    return _jwt_public_key_sign.JwtPublicKeySignInternal

  @core.use_tink_errors
  def primitive(
      self, key_data: tink_pb2.KeyData
  ) -> _jwt_public_key_sign.JwtPublicKeySignInternal:
    sign = self._cc_key_manager.primitive(key_data.SerializeToString())
    algorithm, custom_kid = self._key_data_to_alg_kid(key_data)
    return _JwtPublicKeySign(sign, algorithm, custom_kid)

  def key_type(self) -> str:
    return self._cc_key_manager.key_type()

  @core.use_tink_errors
  def new_key_data(self,
                   key_template: tink_pb2.KeyTemplate) -> tink_pb2.KeyData:
    return tink_pb2.KeyData.FromString(
        self._cc_key_manager.new_key_data(key_template.SerializeToString()))

  @core.use_tink_errors
  def public_key_data(self, key_data: tink_pb2.KeyData) -> tink_pb2.KeyData:
    return tink_pb2.KeyData.FromString(
        self._cc_key_manager.public_key_data(key_data.SerializeToString()))


class _JwtPublicKeyVerifyKeyManagerCcToPyWrapper(
    core.KeyManager[_jwt_public_key_verify.JwtPublicKeyVerifyInternal]):
  """Converts a C++ verify key manager into a JwtPublicKeyVerifyKeyManager."""

  def __init__(self, cc_key_manager: tink_bindings.PublicKeyVerifyKeyManager,
               key_data_to_alg_kid: Callable[[tink_pb2.KeyData],
                                             Tuple[str, Optional[str]]]):
    self._cc_key_manager = cc_key_manager
    self._key_data_to_alg_kid = key_data_to_alg_kid

  def primitive_class(
      self) -> Type[_jwt_public_key_verify.JwtPublicKeyVerifyInternal]:
    return _jwt_public_key_verify.JwtPublicKeyVerifyInternal

  @core.use_tink_errors
  def primitive(
      self, key_data: tink_pb2.KeyData
  ) -> _jwt_public_key_verify.JwtPublicKeyVerifyInternal:
    verify = self._cc_key_manager.primitive(key_data.SerializeToString())
    algorithm, custom_kid = self._key_data_to_alg_kid(key_data)
    return _JwtPublicKeyVerify(verify, algorithm, custom_kid)

  def key_type(self) -> str:
    return self._cc_key_manager.key_type()

  @core.use_tink_errors
  def new_key_data(self,
                   key_template: tink_pb2.KeyTemplate) -> tink_pb2.KeyData:
    return tink_pb2.KeyData.FromString(
        self._cc_key_manager.new_key_data(key_template.SerializeToString()))


def _ecdsa_algorithm_text(algorithm: jwt_ecdsa_pb2.JwtEcdsaAlgorithm) -> str:
  if algorithm not in _ECDSA_ALGORITHM_TEXTS:
    raise _jwt_error.JwtInvalidError('Invalid algorithm')
  return _ECDSA_ALGORITHM_TEXTS[algorithm]


def _get_custom_kid(public_key_proto: Any) -> Optional[str]:
  if public_key_proto.HasField('custom_kid'):
    return public_key_proto.custom_kid.value
  else:
    return None


def _ecdsa_alg_kid_from_private_key_data(
    key_data: tink_pb2.KeyData) -> Tuple[str, Optional[str]]:
  if key_data.type_url != _JWT_ECDSA_PRIVATE_KEY_TYPE:
    raise _jwt_error.JwtInvalidError('Invalid key data key type')
  key = jwt_ecdsa_pb2.JwtEcdsaPrivateKey.FromString(key_data.value)
  return (_ecdsa_algorithm_text(key.public_key.algorithm),
          _get_custom_kid(key.public_key))


def _ecdsa_alg_kid_from_public_key_data(
    key_data: tink_pb2.KeyData) -> Tuple[str, Optional[str]]:
  if key_data.type_url != _JWT_ECDSA_PUBLIC_KEY_TYPE:
    raise _jwt_error.JwtInvalidError('Invalid key data key type')
  key = jwt_ecdsa_pb2.JwtEcdsaPublicKey.FromString(key_data.value)
  return (_ecdsa_algorithm_text(key.algorithm), _get_custom_kid(key))


def _rsa_ssa_pkcs1_algorithm_text(
    algorithm: jwt_rsa_ssa_pkcs1_pb2.JwtRsaSsaPkcs1Algorithm) -> str:
  if algorithm not in _RSA_SSA_PKCS1_ALGORITHM_TEXTS:
    raise _jwt_error.JwtInvalidError('Invalid algorithm')
  return _RSA_SSA_PKCS1_ALGORITHM_TEXTS[algorithm]


def _rsa_ssa_pkcs1_alg_kid_from_private_key_data(
    key_data: tink_pb2.KeyData) -> Tuple[str, Optional[str]]:
  if key_data.type_url != _JWT_RSA_SSA_PKCS1_PRIVATE_KEY_TYPE:
    raise _jwt_error.JwtInvalidError('Invalid key data key type')
  key = jwt_rsa_ssa_pkcs1_pb2.JwtRsaSsaPkcs1PrivateKey.FromString(
      key_data.value)
  return (_rsa_ssa_pkcs1_algorithm_text(key.public_key.algorithm),
          _get_custom_kid(key.public_key))


def _rsa_ssa_pkcs1_alg_kid_from_public_key_data(
    key_data: tink_pb2.KeyData) -> Tuple[str, Optional[str]]:
  if key_data.type_url != _JWT_RSA_SSA_PKCS1_PUBLIC_KEY_TYPE:
    raise _jwt_error.JwtInvalidError('Invalid key data key type')
  key = jwt_rsa_ssa_pkcs1_pb2.JwtRsaSsaPkcs1PublicKey.FromString(key_data.value)
  return (_rsa_ssa_pkcs1_algorithm_text(key.algorithm), _get_custom_kid(key))


def _rsa_ssa_pss_algorithm_text(
    algorithm: jwt_rsa_ssa_pss_pb2.JwtRsaSsaPssAlgorithm) -> str:
  if algorithm not in _RSA_SSA_PSS_ALGORITHM_TEXTS:
    raise _jwt_error.JwtInvalidError('Invalid algorithm')
  return _RSA_SSA_PSS_ALGORITHM_TEXTS[algorithm]


def _rsa_ssa_pss_alg_kid_from_private_key_data(
    key_data: tink_pb2.KeyData) -> Tuple[str, Optional[str]]:
  if key_data.type_url != _JWT_RSA_SSA_PSS_PRIVATE_KEY_TYPE:
    raise _jwt_error.JwtInvalidError('Invalid key data key type')
  key = jwt_rsa_ssa_pss_pb2.JwtRsaSsaPssPrivateKey.FromString(key_data.value)
  return (_rsa_ssa_pss_algorithm_text(key.public_key.algorithm),
          _get_custom_kid(key.public_key))


def _rsa_ssa_pss_alg_kid_from_public_key_data(
    key_data: tink_pb2.KeyData) -> Tuple[str, Optional[str]]:
  if key_data.type_url != _JWT_RSA_SSA_PSS_PUBLIC_KEY_TYPE:
    raise _jwt_error.JwtInvalidError('Invalid key data key type')
  key = jwt_rsa_ssa_pss_pb2.JwtRsaSsaPssPublicKey.FromString(key_data.value)
  return (_rsa_ssa_pss_algorithm_text(key.algorithm), _get_custom_kid(key))


def register():
  """Registers all JWT signature primitives."""
  tink_bindings.register_jwt()

  private_key_manager = _JwtPublicKeySignKeyManagerCcToPyWrapper(
      tink_bindings.PublicKeySignKeyManager.from_cc_registry(
          _JWT_ECDSA_PRIVATE_KEY_TYPE), _ecdsa_alg_kid_from_private_key_data)
  core.Registry.register_key_manager(private_key_manager, new_key_allowed=True)

  public_key_manager = _JwtPublicKeyVerifyKeyManagerCcToPyWrapper(
      tink_bindings.PublicKeyVerifyKeyManager.from_cc_registry(
          _JWT_ECDSA_PUBLIC_KEY_TYPE), _ecdsa_alg_kid_from_public_key_data)
  core.Registry.register_key_manager(public_key_manager, new_key_allowed=True)

  private_key_manager = _JwtPublicKeySignKeyManagerCcToPyWrapper(
      tink_bindings.PublicKeySignKeyManager.from_cc_registry(
          _JWT_RSA_SSA_PKCS1_PRIVATE_KEY_TYPE),
      _rsa_ssa_pkcs1_alg_kid_from_private_key_data)
  core.Registry.register_key_manager(private_key_manager, new_key_allowed=True)

  public_key_manager = _JwtPublicKeyVerifyKeyManagerCcToPyWrapper(
      tink_bindings.PublicKeyVerifyKeyManager.from_cc_registry(
          _JWT_RSA_SSA_PKCS1_PUBLIC_KEY_TYPE),
      _rsa_ssa_pkcs1_alg_kid_from_public_key_data)
  core.Registry.register_key_manager(public_key_manager, new_key_allowed=True)

  private_key_manager = _JwtPublicKeySignKeyManagerCcToPyWrapper(
      tink_bindings.PublicKeySignKeyManager.from_cc_registry(
          _JWT_RSA_SSA_PSS_PRIVATE_KEY_TYPE),
      _rsa_ssa_pss_alg_kid_from_private_key_data)
  core.Registry.register_key_manager(private_key_manager, new_key_allowed=True)

  public_key_manager = _JwtPublicKeyVerifyKeyManagerCcToPyWrapper(
      tink_bindings.PublicKeyVerifyKeyManager.from_cc_registry(
          _JWT_RSA_SSA_PSS_PUBLIC_KEY_TYPE),
      _rsa_ssa_pss_alg_kid_from_public_key_data)
  core.Registry.register_key_manager(public_key_manager, new_key_allowed=True)



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Streaming AEAD wrapper."""

import io
from typing import cast, BinaryIO, Optional, Type

from tink import core
from tink.streaming_aead import _raw_streaming_aead
from tink.streaming_aead import _rewindable_input_stream
from tink.streaming_aead import _streaming_aead


class _DecryptingStreamWrapper(io.RawIOBase):
  """A file-like object which decrypts reads from an underlying object.

  It uses a primitive set of streaming AEADs, and decrypts the stream with the
  matching key in the keyset. Closing this wrapper also closes
  ciphertext_source.
  """

  def __init__(self, primitive_set: core.PrimitiveSet,
               ciphertext_source: BinaryIO, associated_data: bytes):
    """Create a new _DecryptingStreamWrapper.

    Args:
      primitive_set: The primitive set of StreamingAead primitives.
      ciphertext_source: A readable file-like object from which ciphertext bytes
        will be read.
      associated_data: The associated data to use for decryption.
    """
    super().__init__()
    if not ciphertext_source.readable():
      raise ValueError('ciphertext_source must be readable')
    self._ciphertext_source = _rewindable_input_stream.RewindableInputStream(
        ciphertext_source)
    self._associated_data = associated_data
    self._matching_stream = None
    self._remaining_primitives = []
    # For legacy reasons (Tink always encrypted with non-RAW keys) we use all
    # primitives, even those which have output_prefix_type != RAW.
    for entry_list in primitive_set.all():
      for e in entry_list:
        self._remaining_primitives.append(e.primitive)
    self._attempting_stream = self._next_decrypting_stream()

  def _next_decrypting_stream(self) -> io.RawIOBase:
    """Takes the next remaining primitive and returns a decrypting stream."""
    if not self._remaining_primitives:
      raise ValueError('No primitive remaining.')
    # ciphertext_source should never be closed by any of the raw decrypting
    # streams, to be able to use it for another decrypting stream.
    # ciphertext_source will be closed in close().
    # self._ciphertext_source needs to be at the starting position.
    return self._remaining_primitives.pop(0).new_raw_decrypting_stream(
        self._ciphertext_source,
        self._associated_data,
        close_ciphertext_source=False)

  def read(self, size=-1) -> Optional[bytes]:
    """Read and return up to size bytes, where size is an int.

    Args:
      size: Maximum number of bytes to read. As a convenience, if size is
        unspecified or -1, all bytes until EOF are returned.

    Returns:
      Bytes read. An empty bytes object is returned if the stream is already at
      EOF. None is returned if no data is available at the moment.

    Raises:
      TinkError if there was a permanent error.
      ValueError if the file is closed.
    """
    if self.closed:  # pylint:disable=using-constant-test
      raise ValueError('read on closed file.')
    if size == 0:
      return bytes()
    if self._matching_stream:
      return self._matching_stream.read(size)
    # if self._matching_stream is not set, we are currently reading from
    # self._attempting_stream but no data has been read successfully yet.
    while True:
      try:
        if not self._attempting_stream:
          # this should not happen.
          raise ValueError('self._attempting_stream is None')
        data = self._attempting_stream.read(size)
        if data is None:
          # No data at the moment. Not clear if decryption was successful.
          # Try again with the same stream next time.
          return None
        # Any value other than None means that decryption was successful.
        # (b'' indicates that the plaintext is an empty string.)
        self._matching_stream = self._attempting_stream
        self._attempting_stream = None
        self._ciphertext_source.disable_rewind()
        return data
      except core.TinkError as exc:
        if not self._remaining_primitives:
          raise core.TinkError(
              'No matching key found for the ciphertext in the stream') from exc
        # Try another key.
        self._ciphertext_source.rewind()
        self._attempting_stream = self._next_decrypting_stream()

  def readinto(self, b: bytearray) -> Optional[int]:
    """Read bytes into a pre-allocated bytes-like object b."""
    data = self.read(len(b))
    if data is None:
      return None
    n = len(data)
    b[:n] = data
    return n

  def close(self) -> None:
    if self.closed:  # pylint:disable=using-constant-test
      return
    if self._matching_stream:
      self._matching_stream.close()
    if self._attempting_stream:
      self._attempting_stream.close()
    self._ciphertext_source.close()
    super().close()

  def readable(self) -> bool:
    return True


class _WrappedStreamingAead(_streaming_aead.StreamingAead):
  """Implements StreamingAead by wrapping a set of RawStreamingAead."""

  def __init__(self, primitives_set: core.PrimitiveSet):
    self._primitive_set = primitives_set

  def new_encrypting_stream(self, ciphertext_destination: BinaryIO,
                            associated_data: bytes) -> BinaryIO:
    raw = self._primitive_set.primary().primitive.new_raw_encrypting_stream(
        ciphertext_destination, associated_data)
    return cast(BinaryIO, io.BufferedWriter(raw))

  def new_decrypting_stream(self, ciphertext_source: BinaryIO,
                            associated_data: bytes) -> BinaryIO:
    raw = _DecryptingStreamWrapper(self._primitive_set, ciphertext_source,
                                   associated_data)
    return cast(BinaryIO, io.BufferedReader(raw))


class StreamingAeadWrapper(
    core.PrimitiveWrapper[_raw_streaming_aead.RawStreamingAead,
                          _streaming_aead.StreamingAead]):
  """StreamingAeadWrapper is the PrimitiveWrapper for StreamingAead."""

  def wrap(self,
           primitives_set: core.PrimitiveSet) -> _streaming_aead.StreamingAead:
    return _WrappedStreamingAead(primitives_set)

  def primitive_class(self) -> Type[_streaming_aead.StreamingAead]:
    return _streaming_aead.StreamingAead

  def input_primitive_class(
      self) -> Type[_raw_streaming_aead.RawStreamingAead]:
    return _raw_streaming_aead.RawStreamingAead



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for exception (non-)propagation in Pybind11PythonFileObjectAdapter."""

import io

from absl.testing import absltest

import tink
from tink import streaming_aead


def setUpModule():
  streaming_aead.register()


def get_primitive() -> streaming_aead.StreamingAead:
  key_template = streaming_aead.streaming_aead_key_templates.AES128_GCM_HKDF_4KB
  keyset_handle = tink.new_keyset_handle(key_template)
  primitive = keyset_handle.primitive(streaming_aead.StreamingAead)
  return primitive


class BytesIOThatThrowsExceptionsOnReadWrite(io.BytesIO):

  def write(self, data):
    raise tink.TinkError('Called write!')

  def read(self, num):
    raise tink.TinkError('Called read!')

  def close(self):
    pass


class BytesIOThatThrowsExceptionsOnClose(io.BytesIO):

  def write(self, data):
    return len(data)

  def read(self, _):
    return b''

  def close(self):
    raise tink.TinkError('Called close!')


class Pybind11PythonFileObjectAdaterTest(absltest.TestCase):

  # This and the following tests do not use the `with` statement. This is done
  # for two reasons:
  # 1. consistency with the `test_close_throws()`: there, exit from the
  #    context created by the `with` statement causes the `close()` function
  #    to be called after `assertRaises()` verified that it throws -- thus
  #    one more exception is raised, and the test fails.
  # 2. avoiding similar unexpected sideffects in the other tests
  def test_write_throws(self):
    streaming_aead_primitive = get_primitive()

    ciphertext_destination = BytesIOThatThrowsExceptionsOnReadWrite()
    enc_stream = streaming_aead_primitive.new_encrypting_stream(
        ciphertext_destination, b'associated_data')
    # The exception is thrown but swallowed on the way.
    _ = enc_stream.write(b'plaintext')
    # The exception is thrown and is not swallowed.
    self.assertRaises(tink.TinkError, enc_stream.close)

  def test_read_throws(self):
    streaming_aead_primitive = get_primitive()

    ciphertext_source = BytesIOThatThrowsExceptionsOnReadWrite()
    dec_stream = streaming_aead_primitive.new_decrypting_stream(
        ciphertext_source, b'associated_data')
    self.assertRaises(tink.TinkError, dec_stream.read)
    dec_stream.close()

  def test_close_throws(self):
    streaming_aead_primitive = get_primitive()

    ciphertext_destination = BytesIOThatThrowsExceptionsOnClose()
    enc_stream = streaming_aead_primitive.new_encrypting_stream(
        ciphertext_destination, b'associated_data')
    self.assertRaises(tink.TinkError, enc_stream.close)


if __name__ == '__main__':
  absltest.main()



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# NO CHECKED-IN PROTOBUF GENCODE
# source: tink/proto/tink.proto
# Protobuf Python Version: 6.30.2
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import runtime_version as _runtime_version
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
_runtime_version.ValidateProtobufRuntimeVersion(
    _runtime_version.Domain.PUBLIC,
    6,
    30,
    2,
    '',
    'tink/proto/tink.proto'
)
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x15tink/proto/tink.proto\x12\x12google.crypto.tink\"p\n\x0bKeyTemplate\x12\x10\n\x08type_url\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\x0c\x12@\n\x12output_prefix_type\x18\x03 \x01(\x0e\x32$.google.crypto.tink.OutputPrefixType\"\xe8\x01\n\x07KeyData\x12\x10\n\x08type_url\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\x0c\x12\x46\n\x11key_material_type\x18\x03 \x01(\x0e\x32+.google.crypto.tink.KeyData.KeyMaterialType\"t\n\x0fKeyMaterialType\x12\x17\n\x13UNKNOWN_KEYMATERIAL\x10\x00\x12\r\n\tSYMMETRIC\x10\x01\x12\x16\n\x12\x41SYMMETRIC_PRIVATE\x10\x02\x12\x15\n\x11\x41SYMMETRIC_PUBLIC\x10\x03\x12\n\n\x06REMOTE\x10\x04\"\x89\x02\n\x06Keyset\x12\x16\n\x0eprimary_key_id\x18\x01 \x01(\r\x12+\n\x03key\x18\x02 \x03(\x0b\x32\x1e.google.crypto.tink.Keyset.Key\x1a\xb9\x01\n\x03Key\x12-\n\x08key_data\x18\x01 \x01(\x0b\x32\x1b.google.crypto.tink.KeyData\x12\x31\n\x06status\x18\x02 \x01(\x0e\x32!.google.crypto.tink.KeyStatusType\x12\x0e\n\x06key_id\x18\x03 \x01(\r\x12@\n\x12output_prefix_type\x18\x04 \x01(\x0e\x32$.google.crypto.tink.OutputPrefixType\"\x81\x02\n\nKeysetInfo\x12\x16\n\x0eprimary_key_id\x18\x01 \x01(\r\x12\x38\n\x08key_info\x18\x02 \x03(\x0b\x32&.google.crypto.tink.KeysetInfo.KeyInfo\x1a\xa0\x01\n\x07KeyInfo\x12\x10\n\x08type_url\x18\x01 \x01(\t\x12\x31\n\x06status\x18\x02 \x01(\x0e\x32!.google.crypto.tink.KeyStatusType\x12\x0e\n\x06key_id\x18\x03 \x01(\r\x12@\n\x12output_prefix_type\x18\x04 \x01(\x0e\x32$.google.crypto.tink.OutputPrefixType\"`\n\x0f\x45ncryptedKeyset\x12\x18\n\x10\x65ncrypted_keyset\x18\x02 \x01(\x0c\x12\x33\n\x0bkeyset_info\x18\x03 \x01(\x0b\x32\x1e.google.crypto.tink.KeysetInfo*M\n\rKeyStatusType\x12\x12\n\x0eUNKNOWN_STATUS\x10\x00\x12\x0b\n\x07\x45NABLED\x10\x01\x12\x0c\n\x08\x44ISABLED\x10\x02\x12\r\n\tDESTROYED\x10\x03*k\n\x10OutputPrefixType\x12\x12\n\x0eUNKNOWN_PREFIX\x10\x00\x12\x08\n\x04TINK\x10\x01\x12\n\n\x06LEGACY\x10\x02\x12\x07\n\x03RAW\x10\x03\x12\x0b\n\x07\x43RUNCHY\x10\x04\x12\x17\n\x13WITH_ID_REQUIREMENT\x10\x05\x42`\n\x1c\x63om.google.crypto.tink.protoP\x01Z5github.com/tink-crypto/tink-go/v2/proto/tink_go_proto\xa2\x02\x06TINKPBb\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'tink.proto.tink_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  _globals['DESCRIPTOR']._loaded_options = None
  _globals['DESCRIPTOR']._serialized_options = b'\n\034com.google.crypto.tink.protoP\001Z5github.com/tink-crypto/tink-go/v2/proto/tink_go_proto\242\002\006TINKPB'
  _globals['_KEYSTATUSTYPE']._serialized_start=1020
  _globals['_KEYSTATUSTYPE']._serialized_end=1097
  _globals['_OUTPUTPREFIXTYPE']._serialized_start=1099
  _globals['_OUTPUTPREFIXTYPE']._serialized_end=1206
  _globals['_KEYTEMPLATE']._serialized_start=45
  _globals['_KEYTEMPLATE']._serialized_end=157
  _globals['_KEYDATA']._serialized_start=160
  _globals['_KEYDATA']._serialized_end=392
  _globals['_KEYDATA_KEYMATERIALTYPE']._serialized_start=276
  _globals['_KEYDATA_KEYMATERIALTYPE']._serialized_end=392
  _globals['_KEYSET']._serialized_start=395
  _globals['_KEYSET']._serialized_end=660
  _globals['_KEYSET_KEY']._serialized_start=475
  _globals['_KEYSET_KEY']._serialized_end=660
  _globals['_KEYSETINFO']._serialized_start=663
  _globals['_KEYSETINFO']._serialized_end=920
  _globals['_KEYSETINFO_KEYINFO']._serialized_start=760
  _globals['_KEYSETINFO_KEYINFO']._serialized_end=920
  _globals['_ENCRYPTEDKEYSET']._serialized_start=922
  _globals['_ENCRYPTEDKEYSET']._serialized_end=1018
# @@protoc_insertion_point(module_scope)



# Copyright 2023 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for _insecure_keyset_handle."""

from absl.testing import absltest

import tink
from tink import _insecure_keyset_handle
from tink import aead
from tink import core
from tink import secret_key_access
from tink import tink_config


def setUpModule():
  tink_config.register()


class InvalidKeyAccess(core.KeyAccess):
  pass


class InsecureKeysetHandleTest(absltest.TestCase):

  def test_from_to_proto_keyset(self):
    keyset_handle = tink.new_keyset_handle(aead.aead_key_templates.AES128_GCM)
    proto_keyset = _insecure_keyset_handle.to_proto_keyset(
        keyset_handle, secret_key_access.TOKEN
    )
    keyset_handle2 = _insecure_keyset_handle.from_proto_keyset(
        proto_keyset, secret_key_access.TOKEN
    )

    # check that keyset_handle and keyset_handle2 are the same.
    plaintext = b'plaintext'
    associated_data = b'associated_data'
    primitive1 = keyset_handle.primitive(aead.Aead)
    ciphertext = primitive1.encrypt(plaintext, associated_data)
    primitive2 = keyset_handle2.primitive(aead.Aead)
    self.assertEqual(primitive2.decrypt(ciphertext, associated_data), plaintext)

    # check that they fail without secret_key_access.TOKEN
    with self.assertRaises(core.TinkError):
      _insecure_keyset_handle.to_proto_keyset(keyset_handle, InvalidKeyAccess())
    with self.assertRaises(core.TinkError):
      _insecure_keyset_handle.from_proto_keyset(
          proto_keyset, InvalidKeyAccess()
      )


if __name__ == '__main__':
  absltest.main()



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""This module defines KeysetHandle."""

import random
from typing import Optional, Type, TypeVar

from google.protobuf import message
from tink.proto import tink_pb2
from tink import _keyset_reader
from tink import _keyset_writer
from tink import _monitoring
from tink import _secret_key_access
from tink import aead
from tink import core

P = TypeVar('P')

MAX_INT32 = 2147483647  # = 2^31 - 1


class PublicKeyAccess(core.KeyAccess):
  pass


# KeyAccess token that gives access to public keys.
PUBLIC_KEY_ACCESS_TOKEN = PublicKeyAccess()


def has_secret_key_access(token: core.KeyAccess) -> bool:
  """Returns True if token is secret_key_access.TOKEN, and False otherwise."""
  return isinstance(token, _secret_key_access.SecretKeyAccess)


def _validate_output_prefix(
    type_url: str, output_prefix_type: tink_pb2.OutputPrefixType
) -> None:
  """Validates the output prefix type for the given key type."""
  if type_url == 'type.googleapis.com/google.crypto.tink.XAesGcmKey':
    if output_prefix_type not in [
        tink_pb2.TINK,
        tink_pb2.RAW,
    ]:
      raise core.TinkError(
          'invalid output prefix type for XAesGcmKey: {}'.format(
              output_prefix_type
          )
      )


class KeysetHandle:
  """A KeysetHandle provides abstracted access to Keyset.

  KeysetHandle limits the exposure of actual protocol buffers that hold
  sensitive key material. This class allows reading and writing encrypted
  keysets.
  """

  def __new__(cls):
    raise core.TinkError('KeysetHandle cannot be instantiated directly.')

  def __init__(
      self,
      keyset: tink_pb2.Keyset,
      annotations: Optional[_monitoring.Annotations] = None,
  ):
    _validate_keyset(keyset)
    self._keyset = keyset
    self._annotations = annotations

  def __getstate__(self) -> None:
    raise core.TinkError(
        'calling __getstate__ is not permitted. Use'
        ' tink.proto_keyset_format to serialize or deserialize a KeysetHandle.'
    )

  def __setstate__(self, state) -> None:
    _ = state
    raise core.TinkError(
        'calling __setstate__ is not permitted. Use'
        ' tink.proto_keyset_format to serialize or deserialize a KeysetHandle.'
    )

  @classmethod
  def _generate_keyset(
      cls, key_template: tink_pb2.KeyTemplate
  ) -> tink_pb2.Keyset:
    """Returns a single fresh key generated according to key_template."""
    _validate_output_prefix(
        key_template.type_url, key_template.output_prefix_type
    )
    keyset = tink_pb2.Keyset()
    key_data = core.Registry.new_key_data(key_template)
    key_id = _generate_unused_key_id(keyset)
    key = keyset.key.add()
    key.key_data.CopyFrom(key_data)
    key.status = tink_pb2.ENABLED
    key.key_id = key_id
    key.output_prefix_type = key_template.output_prefix_type
    keyset.primary_key_id = key_id
    return keyset

  @classmethod
  def generate_new(cls, key_template: tink_pb2.KeyTemplate) -> 'KeysetHandle':
    """Return a new KeysetHandle.

    It contains a single fresh key generated according to key_template.

    Args:
      key_template: A tink_pb2.KeyTemplate object.

    Returns:
      A new KeysetHandle.
    """
    return cls._create(cls._generate_keyset(key_template))

  @classmethod
  def _generate_new_with_annotations(
      cls,
      key_template: tink_pb2.KeyTemplate,
      annotations: Optional[_monitoring.Annotations],
  ) -> 'KeysetHandle':
    """Same as `generate_new`, but passes along monitoring annotations."""
    keyset = cls._generate_keyset(key_template)
    return cls._create(keyset, annotations)

  @classmethod
  def read(
      cls,
      keyset_reader: _keyset_reader.KeysetReader,
      master_key_aead: aead.Aead,
  ) -> 'KeysetHandle':
    """Tries to create a KeysetHandle from an encrypted keyset."""
    return cls.read_with_associated_data(keyset_reader, master_key_aead, b'')

  @classmethod
  def read_with_associated_data(
      cls,
      keyset_reader: _keyset_reader.KeysetReader,
      master_key_aead: aead.Aead,
      associated_data: bytes,
  ) -> 'KeysetHandle':
    """Tries to create a KeysetHandle from an encrypted keyset using the provided associated data."""
    encrypted_keyset = keyset_reader.read_encrypted()
    _assert_enough_encrypted_key_material(encrypted_keyset)
    return cls._create(
        _decrypt(encrypted_keyset, master_key_aead, associated_data)
    )

  @classmethod
  def read_no_secret(
      cls, keyset_reader: _keyset_reader.KeysetReader
  ) -> 'KeysetHandle':
    """Creates a KeysetHandle from a keyset with no secret key material.

    This can be used to load public keysets or envelope encryption keysets.

    Args:
      keyset_reader: A _keyset_reader.KeysetReader object.

    Returns:
      A new KeysetHandle.
    """
    keyset = keyset_reader.read()
    _assert_no_secret_key_material(keyset)
    return cls._create(keyset)

  @classmethod
  def _create(
      cls,
      keyset: tink_pb2.Keyset,
      annotations: Optional[_monitoring.Annotations] = None,
  ):
    o = object.__new__(cls)
    o.__init__(keyset, annotations)
    return o

  def keyset_info(self) -> tink_pb2.KeysetInfo:
    """Returns the KeysetInfo that doesn't contain actual key material."""
    return _keyset_info(self._keyset)

  def write(
      self,
      keyset_writer: _keyset_writer.KeysetWriter,
      master_key_primitive: aead.Aead,
  ) -> None:
    """Serializes, encrypts with master_key_primitive and writes the keyset."""
    self.write_with_associated_data(keyset_writer, master_key_primitive, b'')

  def write_with_associated_data(
      self,
      keyset_writer: _keyset_writer.KeysetWriter,
      master_key_primitive: aead.Aead,
      associated_data: bytes,
  ) -> None:
    """Serializes, encrypts with master_key_primitive and writes the keyset."""
    encrypted_keyset = _encrypt(
        self._keyset, master_key_primitive, associated_data
    )
    keyset_writer.write_encrypted(encrypted_keyset)

  def write_no_secret(self, keyset_writer: _keyset_writer.KeysetWriter) -> None:
    """Writes the underlying keyset to keyset_writer.

    Writes the underlying keyset to keyset_writer only if the keyset does not
    contain any secret key material.
    This can be used to persist public keysets or envelope encryption keysets.
    Users that need to persist keysets with secret material can use
    cleartext_keyset_handle.

    Args:
      keyset_writer: A KeysetWriter object.
    """
    _assert_no_secret_key_material(self._keyset)
    keyset_writer.write(self._keyset)

  def public_keyset_handle(self) -> 'KeysetHandle':
    """Returns a new KeysetHandle for the corresponding public keys."""
    public_keyset = tink_pb2.Keyset()
    for key in self._keyset.key:
      public_key = public_keyset.key.add()
      public_key.CopyFrom(key)
      public_key.key_data.CopyFrom(core.Registry.public_key_data(key.key_data))
      _validate_key(public_key)
    public_keyset.primary_key_id = self._keyset.primary_key_id
    return self._create(public_keyset, self._annotations)

  def primitive(self, primitive_class: Type[P]) -> P:
    """Returns a wrapped primitive from this KeysetHandle.

    Uses the KeyManager and the PrimitiveWrapper objects in the global
    registry.Registry
    to create the primitive. This function is the most common way of creating a
    primitive.

    Args:
      primitive_class: The class of the primitive.

    Returns:
      The primitive.
    Raises:
      tink.TinkError if creation of the primitive fails, for example if
      primitive_class cannot be used with this KeysetHandle.
    """
    _validate_keyset(self._keyset)
    input_primitive_class = core.Registry.input_primitive_class(primitive_class)
    pset = core.PrimitiveSet(input_primitive_class)
    for key in self._keyset.key:
      if key.status == tink_pb2.ENABLED:
        _validate_output_prefix(key.key_data.type_url, key.output_prefix_type)
        primitive = core.Registry.primitive(key.key_data, input_primitive_class)
        entry = pset.add_primitive(primitive, key)
        if key.key_id == self._keyset.primary_key_id:
          pset.set_primary(entry)
    # pylint: disable-next=protected-access
    return core.Registry._wrap_with_monitoring_info(
        pset,
        primitive_class,
        _monitoring.MonitoringKeySetInfo(self._annotations, self.keyset_info()),
    )


def new_keyset_handle(key_template: tink_pb2.KeyTemplate) -> KeysetHandle:
  return _new_keyset_handle_with_annotations(key_template, None)


def _new_keyset_handle_with_annotations(
    key_template: tink_pb2.KeyTemplate,
    annotations: Optional[_monitoring.Annotations],
) -> KeysetHandle:
  # pylint: disable-next=protected-access
  keyset_handle = KeysetHandle._generate_new_with_annotations(
      key_template, annotations
  )
  return keyset_handle


def read_keyset_handle(
    keyset_reader: _keyset_reader.KeysetReader, master_key_aead: aead.Aead
) -> KeysetHandle:
  return KeysetHandle.read(keyset_reader, master_key_aead)


def read_keyset_handle_with_associated_data(
    keyset_reader: _keyset_reader.KeysetReader,
    master_key_aead: aead.Aead,
    associated_data: bytes,
) -> KeysetHandle:
  return KeysetHandle.read_with_associated_data(
      keyset_reader, master_key_aead, associated_data
  )


def read_no_secret_keyset_handle(
    keyset_reader: _keyset_reader.KeysetReader,
) -> KeysetHandle:
  return KeysetHandle.read_no_secret(keyset_reader)


def _keyset_info(keyset: tink_pb2.Keyset) -> tink_pb2.KeysetInfo:
  keyset_info = tink_pb2.KeysetInfo(primary_key_id=keyset.primary_key_id)
  for key in keyset.key:
    key_info = keyset_info.key_info.add()
    key_info.type_url = key.key_data.type_url
    key_info.status = key.status
    key_info.output_prefix_type = key.output_prefix_type
    key_info.key_id = key.key_id
  return keyset_info


def _encrypt(
    keyset: tink_pb2.Keyset,
    master_key_primitive: aead.Aead,
    associated_data: bytes,
) -> tink_pb2.EncryptedKeyset:
  """Encrypts a Keyset and returns an EncryptedKeyset."""
  encrypted_keyset = master_key_primitive.encrypt(
      keyset.SerializeToString(), associated_data
  )
  # Check if we can decrypt, to detect errors
  try:
    keyset2 = tink_pb2.Keyset.FromString(
        master_key_primitive.decrypt(encrypted_keyset, associated_data)
    )
    if keyset != keyset2:
      raise core.TinkError(
          'cannot encrypt keyset: %s != %s' % (keyset, keyset2)
      )
  except message.DecodeError:
    raise core.TinkError('invalid keyset, corrupted key material')
  return tink_pb2.EncryptedKeyset(
      encrypted_keyset=encrypted_keyset, keyset_info=_keyset_info(keyset)
  )


def _decrypt(
    encrypted_keyset: tink_pb2.EncryptedKeyset,
    master_key_aead: aead.Aead,
    associated_data: bytes,
) -> tink_pb2.Keyset:
  """Decrypts an EncryptedKeyset and returns a Keyset."""
  try:
    keyset = tink_pb2.Keyset.FromString(
        master_key_aead.decrypt(
            encrypted_keyset.encrypted_keyset, associated_data
        )
    )
    # Check emptiness here too, in case the encrypted keys unwrapped to nothing?
    _assert_enough_key_material(keyset)
    return keyset
  except message.DecodeError:
    raise core.TinkError('invalid keyset, corrupted key material')


def _validate_keyset(keyset: tink_pb2.Keyset):
  """Raises tink_error.TinkError if keyset is not valid."""
  for key in keyset.key:
    if key.status != tink_pb2.DESTROYED:
      _validate_key(key)
  num_non_destroyed_keys = sum(
      1 for key in keyset.key if key.status != tink_pb2.DESTROYED
  )
  num_non_public_key_material = sum(
      1
      for key in keyset.key
      if key.key_data.key_material_type != tink_pb2.KeyData.ASYMMETRIC_PUBLIC
  )
  num_primary_keys = sum(
      1
      for key in keyset.key
      if key.status == tink_pb2.ENABLED and key.key_id == keyset.primary_key_id
  )
  if num_non_destroyed_keys == 0:
    raise core.TinkError('empty keyset')
  if num_primary_keys > 1:
    raise core.TinkError('keyset contains multiple primary keys')
  if num_primary_keys == 0 and num_non_public_key_material > 0:
    raise core.TinkError('keyset does not contain a valid primary key')


def _validate_key(key: tink_pb2.Keyset.Key):
  """Raises tink_error.TinkError if key is not valid."""
  if not key.HasField('key_data'):
    raise core.TinkError('key {} has no key data'.format(key.key_id))
  if key.output_prefix_type == tink_pb2.UNKNOWN_PREFIX:
    raise core.TinkError('key {} has unknown prefix'.format(key.key_id))
  if key.status == tink_pb2.UNKNOWN_STATUS:
    raise core.TinkError('key {} has unknown status'.format(key.key_id))


def _assert_no_secret_key_material(keyset: tink_pb2.Keyset):
  for key in keyset.key:
    if key.key_data.key_material_type in (
        tink_pb2.KeyData.UNKNOWN_KEYMATERIAL,
        tink_pb2.KeyData.SYMMETRIC,
        tink_pb2.KeyData.ASYMMETRIC_PRIVATE,
    ):
      raise core.TinkError('keyset contains secret key material')


def _assert_enough_key_material(keyset: tink_pb2.Keyset):
  if not keyset or not keyset.key:
    raise core.TinkError('empty keyset')


def _assert_enough_encrypted_key_material(
    encrypted_keyset: tink_pb2.EncryptedKeyset,
):
  if not encrypted_keyset or not encrypted_keyset.encrypted_keyset:
    raise core.TinkError('empty keyset')


def _generate_unused_key_id(keyset: tink_pb2.Keyset) -> int:
  while True:
    key_id = random.randint(1, MAX_INT32)
    if key_id not in {key.key_id for key in keyset.key}:
      return key_id



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""CleartextKeysetHandle module.

WARNING

Reading or writing cleartext keysets is a bad practice, usage of this API
should be restricted. Users can read encrypted keysets using
tink.read_keyset_handle.
"""

from tink.proto import tink_pb2
import tink


def from_keyset(keyset: tink_pb2.Keyset) -> tink.KeysetHandle:
  """Create a KeysetHandle from a keyset."""
  return tink.KeysetHandle._create(keyset)  # pylint: disable=protected-access


def read(keyset_reader: tink.KeysetReader) -> tink.KeysetHandle:
  """Create a KeysetHandle from a keyset_reader."""
  keyset = keyset_reader.read()
  return tink.KeysetHandle._create(keyset)  # pylint: disable=protected-access


def write(keyset_writer: tink.KeysetWriter,
          keyset_handle: tink.KeysetHandle) -> None:
  """Serializes and writes the keyset."""
  keyset_writer.write(keyset_handle._keyset)  # pylint: disable=protected-access



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# NO CHECKED-IN PROTOBUF GENCODE
# source: tink/proto/test_proto.proto
# Protobuf Python Version: 6.30.2
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import runtime_version as _runtime_version
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
_runtime_version.ValidateProtobufRuntimeVersion(
    _runtime_version.Domain.PUBLIC,
    6,
    30,
    2,
    '',
    'tink/proto/test_proto.proto'
)
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x1btink/proto/test_proto.proto\x12\x12google.crypto.tink\"%\n\tTestProto\x12\x0b\n\x03num\x18\x01 \x01(\x04\x12\x0b\n\x03str\x18\x02 \x01(\x0c\"\x7f\n\x0fNestedTestProto\x12(\n\x01\x61\x18\x01 \x01(\x0b\x32\x1d.google.crypto.tink.TestProto\x12(\n\x01\x62\x18\x02 \x01(\x0b\x32\x1d.google.crypto.tink.TestProto\x12\x0b\n\x03num\x18\x03 \x01(\x04\x12\x0b\n\x03str\x18\x04 \x01(\x0c\"$\n\x15TestProtoWithoutCtype\x12\x0b\n\x03str\x18\x01 \x01(\x0c\x42 \n\x1c\x63om.google.crypto.tink.protoP\x01\x62\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'tink.proto.test_proto_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  _globals['DESCRIPTOR']._loaded_options = None
  _globals['DESCRIPTOR']._serialized_options = b'\n\034com.google.crypto.tink.protoP\001'
  _globals['_TESTPROTO']._serialized_start=51
  _globals['_TESTPROTO']._serialized_end=88
  _globals['_NESTEDTESTPROTO']._serialized_start=90
  _globals['_NESTEDTESTPROTO']._serialized_end=217
  _globals['_TESTPROTOWITHOUTCTYPE']._serialized_start=219
  _globals['_TESTPROTOWITHOUTCTYPE']._serialized_end=255
# @@protoc_insertion_point(module_scope)
